
# Imports
import os
import random

from ...decorators import handle_error
from ..config.get import DataScienceConfig
from .image_preprocess import ImageDatasetPreprocess
from .technique import ProcessingTechnique


# Image dataset augmentation class
class ImageDatasetAugmentation(ImageDatasetPreprocess):
	""" Image dataset augmentation class. Check the class constructor for more information. """

	# Class constructor (configuration)
	def __init__(self, final_dataset_size: int, techniques: list[ProcessingTechnique] | None = None) -> None:
		""" Initialize the image dataset augmentation class with the given parameters.

		Args:
			final_dataset_size	(int):							Size of the final dataset
			techniques			(list[ProcessingTechnique]):	List of processing techniques to apply.
		"""
		if techniques is None:
			techniques = []
		super().__init__(techniques=techniques)
		self.final_dataset_size: int = final_dataset_size

	# Class methods
	def _add_suffix(self, path: str, used_destinations: set[str]) -> str:
		""" Add a count suffix to the path in order to avoid overwriting the same file

		Args:
			path	(str):	Path to the file (example: "path/to/file.jpg")
		Returns:
			str:	Path with the suffix (example: "path/to/file_1.jpg")
		"""
		# Split the path into base path and extension (e.g. "path/to/file.jpg" -> "path/to/file", ".jpg")
		path_no_ext, ext = os.path.splitext(path)

		# Convert count to augmented path
		def get_path_from_count(count: int) -> str:
			return path_no_ext + DataScienceConfig.AUGMENTED_FILE_SUFFIX + str(count) + ext

		# Function to check if the path is not available
		def is_not_available(path_aug: str) -> bool:
			return path_aug in used_destinations or os.path.isfile(path_aug)

		# Keep incrementing counter until we find a filename that doesn't exist
		count: int = 1
		while is_not_available(get_path_from_count(count)):
			count += 1
		return get_path_from_count(count)

	@handle_error(message="Error while getting queue of files to process")
	def get_queue(
		self,
		dataset_path: str,
		destination_path: str,
		images_per_class_dict: dict[str, int] | None = None
	) -> list[tuple[str, str, list[ProcessingTechnique]]]:
		""" Get the queue of images to process with their techniques.

		Args:
			dataset_path          (str):             Path to the dataset
			destination_path      (str):             Path to the destination dataset
			images_per_class_dict (dict[str, int]):  Dictionary mapping class names to desired number of images
				(optional, defaults to empty dictionary)
		Returns:
			list[tuple[str, str, list[ProcessingTechnique]]]: Queue of (source_path, dest_path, techniques) tuples
		"""
		# Initializations
		if images_per_class_dict is None:
			images_per_class_dict = {}
		queue: list[tuple[str, str, list[ProcessingTechnique]]] = []
		used_destinations: set[str] = set()

		# Get all folders (classes) and compute the number of images per class
		classes: tuple[str, ...] = tuple(f for f in os.listdir(dataset_path) if os.path.isdir(f"{dataset_path}/{f}"))
		default_images_per_class: int = self.final_dataset_size // len(classes)

		# For each class, for each image, apply the processing techniques
		for class_name in classes:
			class_path: str = f"{dataset_path}/{class_name}"
			images: list[str] = os.listdir(class_path)

			# Determine target number of images for this class
			target_images: int = images_per_class_dict.get(class_name, default_images_per_class)
			remaining_images: int = target_images - len(images)

			# Add images to the queue without applying the processing techniques
			for img in images:
				files: dict[str, str] = self.get_files_recursively(f"{class_path}/{img}", f"{destination_path}/{class_name}/{img}")
				for path, dest in files.items():
					queue.append((path, dest, []))

			# While there is less images than the desired number, apply the processing techniques
			while remaining_images > 0:
				chosen_images: list[str] = random.sample(images, k=min(remaining_images, len(images)))

				# Apply the processing techniques
				for img in chosen_images:
					img_destination: str = self._add_suffix(f"{destination_path}/{class_name}/{img}", used_destinations)
					used_destinations.add(img_destination)
					img_path: str = f"{class_path}/{img}"

					# Get the technique and their fixed values
					techniques: list[ProcessingTechnique] = [x.deterministic(use_default=False) for x in self.techniques]

					# For each image found, add it to the queue
					for path, dest in self.get_files_recursively(img_path, img_destination).items():
						queue.append((path, dest, techniques))

				# Update the remaining images
				remaining_images -= len(chosen_images)

		return queue




from ..decorators import LogLevels as LogLevels, handle_error as handle_error
from collections.abc import Callable as Callable

def generate_stubs(package_name: str, extra_args: str = '--include-docstrings --include-private') -> int:
    ''' Generate stub files for a Python package using stubgen.

\tNote: stubgen generates stubs in the \'out\' directory by default in the current working directory.

\tArgs:
\t\tpackage_name  (str): Name of the package to generate stubs for.
\t\textra_args    (str): Extra arguments to pass to stubgen. Defaults to "--include-docstrings --include-private".
\tReturns:
\t\tint: Return code of the os.system call.
\t'''
def clean_stubs_directory(output_directory: str, package_name: str) -> None:
    """ Clean the stubs directory by deleting all .pyi files.

\tArgs:
\t\toutput_directory  (str): Directory to clean.
\t\tpackage_name      (str): Package name subdirectory. Only cleans output_directory/package_name.
\t"""
def stubs_full_routine(package_name: str, output_directory: str = 'typings', extra_args: str = '--include-docstrings --include-private', clean_before: bool = False, generate_stubs_function: Callable[[str, str], int] = ..., clean_stubs_function: Callable[[str, str], None] = ...) -> None:
    ''' Generate stub files for a Python package using stubgen.

\tNote: stubgen generates stubs in the \'out\' directory by default in the current working directory.

\tArgs:
\t\tpackage_name              (str):                       Name of the package to generate stubs for.
\t\toutput_directory          (str):                       Directory to clean before generating stubs. Defaults to "typings".
\t\t\tThis parameter is used for cleaning the directory before stub generation.
\t\textra_args                (str):                       Extra arguments to pass to stubgen. Defaults to "--include-docstrings --include-private".
\t\tclean_before              (bool):                      Whether to clean the output directory before generating stubs. Defaults to False.
\t\tgenerate_stubs_function   (Callable[[str, str], int]): Function to generate stubs.
\t\t\tDefaults to :func:`generate_stubs`.
\t\tclean_stubs_function      (Callable[[str], None]):     Function to clean the stubs directory.
\t\t\tDefaults to :func:`clean_stubs_directory`.
\tRaises:
\t\tException: If stub generation fails.
\t'''





# PYTHON_ARGCOMPLETE_OK
# Imports
import argparse
import sys

import argcomplete

from .all_doctests import launch_tests
from .archive import archive_cli
from .backup import backup_cli
from .decorators import handle_error
from .print import CYAN, GREEN, RESET
from .version_pkg import show_version_cli

# Argument Parser Setup for Auto-Completion
parser = argparse.ArgumentParser(prog="stouputils", add_help=False)
parser.add_argument("command", nargs="?", choices=[
	"--version", "-v", "version", "show_version", "all_doctests", "archive", "backup"
])
parser.add_argument("args", nargs="*")
argcomplete.autocomplete(parser)


@handle_error(message="Error while running 'stouputils'")
def main() -> None:
	second_arg: str = sys.argv[1].lower() if len(sys.argv) >= 2 else ""

	# Print the version of stouputils and its dependencies
	if second_arg in ("--version", "-v", "version", "show_version"):
		return show_version_cli()

	# Handle "all_doctests" command
	if second_arg == "all_doctests":
		if launch_tests("." if len(sys.argv) == 2 else sys.argv[2]) > 0:
			sys.exit(1)
		return

	# Handle "archive" command
	if second_arg == "archive":
		sys.argv.pop(1)  # Remove "archive" from argv so archive_cli gets clean arguments
		return archive_cli()

	# Handle "backup" command
	if second_arg == "backup":
		sys.argv.pop(1)  # Remove "backup" from argv so backup_cli gets clean arguments
		return backup_cli()

	# Check if the command is any package name
	if second_arg in (): # type: ignore
		return

	# Get version
	from importlib.metadata import version
	try:
		pkg_version = version("stouputils")
	except Exception:
		pkg_version = "unknown"

	# Print help with nice formatting
	separator: str = "─" * 60
	print(f"""
{CYAN}{separator}{RESET}
{CYAN}stouputils {GREEN}CLI {CYAN}v{pkg_version}{RESET}
{CYAN}{separator}{RESET}
{CYAN}Usage:{RESET} stouputils <command> [options]

{CYAN}Available commands:{RESET}
  {GREEN}--version, -v{RESET} [pkg] [-t <depth>]
        Show version information (optionally for a specific package)
  {GREEN}all_doctests{RESET} [dir]   Run all doctests in the specified directory
  {GREEN}archive{RESET} --help       Archive utilities (make, repair)
  {GREEN}backup{RESET} --help        Backup utilities (delta, consolidate, limit)
{CYAN}{separator}{RESET}
""".strip())
	return

if __name__ == "__main__":
	main()




""" MobileNet models implementation.

This module provides wrapper classes for the MobileNet family of models from the Keras applications.
MobileNet models are designed for mobile and embedded vision applications,
offering efficient architectures that deliver high accuracy with low computational requirements.

Available models:

- MobileNet: Original MobileNet architecture using depthwise separable convolutions
- MobileNetV2: Lightweight architecture using inverted residuals and linear bottlenecks
- MobileNetV3Small: Compact variant of MobileNetV3 optimized for mobile devices
- MobileNetV3Large: Larger variant of MobileNetV3 with higher capacity

All models support transfer learning from ImageNet pre-trained weights.
"""
# pyright: reportUnknownVariableType=false
# pyright: reportMissingTypeStubs=false

# Imports
from __future__ import annotations

from keras.models import Model
from keras.src.applications.mobilenet import MobileNet as MobileNet_keras
from keras.src.applications.mobilenet_v2 import MobileNetV2 as MobileNetV2_keras
from keras.src.applications.mobilenet_v3 import MobileNetV3Large as MobileNetV3Large_keras
from keras.src.applications.mobilenet_v3 import MobileNetV3Small as MobileNetV3Small_keras

from ....decorators import simple_cache
from ..base_keras import BaseKeras
from ..model_interface import CLASS_ROUTINE_DOCSTRING, MODEL_DOCSTRING


# Classes
class MobileNet(BaseKeras):
	def _get_base_model(self) -> Model:
		return MobileNet_keras(include_top=False, classes=self.num_classes)

class MobileNetV2(BaseKeras):
	def _get_base_model(self) -> Model:
		return MobileNetV2_keras(include_top=False, classes=self.num_classes)

class MobileNetV3Small(BaseKeras):
	def _get_base_model(self) -> Model:
		return MobileNetV3Small_keras(include_top=False, classes=self.num_classes)

class MobileNetV3Large(BaseKeras):
	def _get_base_model(self) -> Model:
		return MobileNetV3Large_keras(include_top=False, classes=self.num_classes)


# Docstrings
for model in [MobileNet, MobileNetV2, MobileNetV3Small, MobileNetV3Large]:
	model.__doc__ = MODEL_DOCSTRING.format(model=model.__name__)
	model.class_routine = simple_cache(model.class_routine)
	model.class_routine.__doc__ = CLASS_ROUTINE_DOCSTRING.format(model=model.__name__)




"""
This module contains the Dataset class, which provides an easy way to handle ML datasets.

The Dataset class has the following attributes:

- training_data        (XyTuple):              Training data containing features, labels and file paths
- test_data            (XyTuple):              Test data containing features, labels and file paths
- num_classes          (int):                  Number of classes in the dataset
- name                 (str):                  Name of the dataset
- grouping_strategy    (GroupingStrategy):     Strategy for grouping images when loading
- labels               (list[str]):            List of class labels (strings)
- loading_type         (Literal["image"]):     Type of the dataset (currently only "image" is supported)
- original_dataset     (Dataset | None):       Original dataset used for data augmentation
- class_distribution   (dict[str, dict]):      Class distribution counts for train/test sets

It provides methods for:

- Loading image datasets from directories using different grouping strategies
- Splitting data into train/test sets with stratification (and care for data augmentation)
- Managing class distributions and dataset metadata
"""
# pyright: reportUnknownMemberType=false

# Imports
from __future__ import annotations

import os
from collections.abc import Generator, Iterable
from typing import Any, Literal

import numpy as np
from numpy.typing import NDArray

from ...decorators import handle_error, LogLevels
from ...print import warning, progress
from ...collections import unique_list
from ..utils import Utils
from .grouping_strategy import GroupingStrategy
from .xy_tuple import XyTuple

# Constants
DEFAULT_IMAGE_KWARGS: dict[str, Any] = {
	"image_size": (224, 224),
	"label_mode": "categorical",
	"color_mode": "rgb",
	"batch_size": 1
}
""" Default image kwargs sent to keras.image_dataset_from_directory """

# Dataset class
class Dataset:
	""" Dataset class used for easy data handling. """

	# Class constructors
	def __init__(
		self,
		training_data: XyTuple | list[Any],
		val_data: XyTuple | list[Any] | None = None,
		test_data: XyTuple | list[Any] | None = None,
		name: str = "",
		grouping_strategy: GroupingStrategy = GroupingStrategy.NONE,
		labels: tuple[str, ...] = (),
		loading_type: Literal["image"] = "image"
	) -> None:
		""" Initialize the dataset class

		>>> Dataset(training_data=tuple(), test_data=tuple(), name="doctest")
		Traceback (most recent call last):
			...
		AssertionError: data must be a tuple with X and y as iterables
		"""
		if val_data is None:
			val_data = XyTuple.empty()
		if test_data is None:
			test_data = XyTuple.empty()

		# Assertions
		all_data: tuple[Any, ...] = (training_data, val_data, test_data)
		for data in all_data:
			if not isinstance(data, XyTuple):
				assert isinstance(data, Iterable) \
					and 2 <= len(data) <= 3 \
					and isinstance(data[0], Iterable) \
					and isinstance(data[1], Iterable), "data must be a tuple with X and y as iterables"

		# Get training, validation and test data
		xy_tuples: list[XyTuple] = [XyTuple(*data) if not isinstance(data, XyTuple) else data for data in all_data]

		# Pre-process for attributes initialization
		num_classes: int = self._get_num_classes(xy_tuples[0].y, xy_tuples[1].y, xy_tuples[2].y)
		labels = tuple(str(x).replace("_", " ").title() for x in (labels if labels else range(num_classes)))

		# Initialize attributes
		self._training_data: XyTuple = xy_tuples[0]
		""" Training data as XyTuple containing X and y as numpy arrays.
		This is a protected attribute accessed via the public property self.training_data. """
		self._val_data: XyTuple = xy_tuples[1]
		""" Validation data as XyTuple containing X and y as numpy arrays.
		This is a protected attribute accessed via the public property self.val_data. """
		self._test_data: XyTuple = xy_tuples[2]
		""" Test data as XyTuple containing X and y as numpy arrays.
		This is a protected attribute accessed via the public property self.test_data. """
		self.num_classes: int = num_classes
		""" Number of classes in the dataset (y) """
		self.name: str = os.path.basename(name)
		""" Name of the dataset (path given in the constructor are converted,
		ex: ".../data/pizza_not_pizza" becomes "pizza_not_pizza") """
		self.loading_type: Literal["image"] = loading_type
		""" Type of the dataset """
		self.grouping_strategy: GroupingStrategy = grouping_strategy
		""" Grouping strategy for the dataset """
		self.labels: tuple[str, ...] = labels
		""" List of class labels (strings) """
		self.class_distribution: dict[str, dict[int, int]] = {"train": {}, "val": {}, "test": {}}
		""" Class distribution in the dataset for both training and test sets """
		self.original_dataset: Dataset | None = None
		""" Original dataset used for data augmentation (can be None) """

		# Update class distribution
		self._update_class_distribution()

	def _get_num_classes(self, *values: Any) -> int:
		""" Get the number of classes in the dataset.

		Args:
			values (NDArray[Any]): Arrays containing class labels
		Returns:
			int: Number of unique classes
		"""
		# Handle case where arrays have different dimensions (1D vs 2D)
		processed_values: list[NDArray[Any]] = []
		for value in values:
			value: NDArray[Any] = np.array(value)
			if len(value.shape) == 2:  # One-hot encoded
				processed_values.append(Utils.convert_to_class_indices(value))
			else:
				processed_values.append(value)

		return len(np.unique(np.concatenate(processed_values)))

	def _update_class_distribution(self, update_num_classes: bool = False) -> None:
		""" Update the class distribution dictionary for both training and test data. """
		# For each data type,
		for data_type, data in (("train", self._training_data), ("val", self._val_data), ("test", self._test_data)):
			y_data: NDArray[Any] = np.array(data.y)
			if len(y_data.shape) == 2:  # One-hot encoded
				y_data = Utils.convert_to_class_indices(y_data)

			# Update the class distribution
			self.class_distribution[data_type] = {}
			for class_id in range(self.num_classes):
				self.class_distribution[data_type][class_id] = np.sum(y_data == class_id)

		# Update the number of classes if needed
		if update_num_classes:
			self.num_classes = self._get_num_classes(self._training_data.y, self._val_data.y, self._test_data.y)

	def exclude_augmented_images_from_val_test(self, original_dataset: Dataset) -> None:
		""" Exclude augmented versions of validation and test images from the training set.

		This ensures that augmented versions of images in the validation and test sets are not present in the training set,
		which would cause data leakage.

		Args:
			original_dataset (Dataset): The original dataset containing the test images to exclude
		"""
		# Get base filenames from original test set
		progress("Excluding augmented versions of validation and test images from training set...")
		val_test_base_names: list[list[str]] = [
			[os.path.splitext(os.path.basename(f))[0] for f in filepaths]
			for filepaths in (*original_dataset.val_data.filepaths, *original_dataset.test_data.filepaths)
		]
		val_test_base_names = unique_list(val_test_base_names, method="str")

		# Get base filenames from training set
		train_base_names: list[list[str]] = [
			[os.path.splitext(os.path.basename(f))[0] for f in filepaths]
			for filepaths in self.training_data.filepaths
		]

		# Remove augmented versions of test images from training set
		# Get indices of training samples that are not augmented versions of test samples
		# For each training sample, check if any of its filenames start with any test filename
		train_indices: list[int] = [
			i for i, train_names in enumerate(train_base_names)
			if not any(
				any(train_name.startswith(name) for train_name in train_names)
				for names in val_test_base_names
				for name in names
			)
		]

		# Update training data to exclude augmented versions
		self._training_data = XyTuple(
			[self.training_data.X[i] for i in train_indices],
			[self.training_data.y[i] for i in train_indices],
			tuple(self.training_data.filepaths[i] for i in train_indices)
		)

		# Use original test data
		self._test_data = original_dataset.test_data     # Impossible to have augmented test_data here
		self._val_data = original_dataset.val_data       # Impossible to have augmented val_data here
		self._update_class_distribution(update_num_classes=False)

	@property
	def training_data(self) -> XyTuple:
		return self._training_data

	@training_data.setter
	def training_data(self, value: XyTuple | Any) -> None:
		warning("Setting training data...", value)
		self._training_data = XyTuple(*value) if not isinstance(value, XyTuple) else value
		self._update_class_distribution(update_num_classes=True)

	@property
	def val_data(self) -> XyTuple:
		return self._val_data

	@val_data.setter
	def val_data(self, value: XyTuple | Any) -> None:
		self._val_data = XyTuple(*value) if not isinstance(value, XyTuple) else value
		self._update_class_distribution(update_num_classes=True)

	@property
	def test_data(self) -> XyTuple:
		return self._test_data

	@test_data.setter
	def test_data(self, value: XyTuple | Any) -> None:
		self._test_data = XyTuple(*value) if not isinstance(value, XyTuple) else value
		self._update_class_distribution(update_num_classes=True)

	# Class methods
	def __str__(self) -> str:
		train_dist: dict[int, int] = self.class_distribution["train"]
		val_dist: dict[int, int] = self.class_distribution["val"]
		test_dist: dict[int, int] = self.class_distribution["test"]
		return (
			f"Dataset {self.name}: "
			f"{len(self.training_data.X):,} training samples, "
			f"{len(self.val_data.X):,} validation samples, "
			f"{len(self.test_data.X):,} test samples, "
			f"{self.num_classes:,} classes "
			f"(train: {train_dist}, val: {val_dist}, test: {test_dist})"
		)

	def __repr__(self) -> str:
		return (
			f"Dataset(training_data={self.training_data!r}, "
			f"val_data={self.val_data!r}, "
			f"test_data={self.test_data!r}, "
			f"num_classes={self.num_classes}, "
			f"name={self.name!r}, "
			f"grouping_strategy={self.grouping_strategy.name})"
		)

	def __iter__(self) -> Generator[XyTuple, Any, Any]:
		""" Allow unpacking of the dataset into train and test sets.

		Returns:
			Generator[XyTuple], Any, Any]: Generator over the dataset splits

		>>> X, y = [[1]], [2]
		>>> dataset = Dataset(training_data=(X, y), test_data=(X, y), name="doctest")
		>>> train, val, test = dataset
		>>> train == (X, y) and val == () and test == (X, y)
		True
		>>> train == XyTuple(X, y) and val == XyTuple.empty() and test == XyTuple(X, y)
		True
		"""
		yield from (self.training_data, self.val_data, self.test_data)

	def get_experiment_name(self, override_name: str = "") -> str:
		""" Get the experiment name for mlflow, e.g. "DatasetName_GroupingStrategyName"

		Args:
			override_name (str): Override the Dataset name
		Returns:
			str: Experiment name
		"""
		if override_name:
			return f"{override_name}_{self.grouping_strategy.name.title()}"
		else:
			return f"{self.name}_{self.grouping_strategy.name.title()}"


	# Static methods
	@staticmethod
	@handle_error(error_log=LogLevels.ERROR_TRACEBACK)
	def empty() -> Dataset:
		return Dataset(XyTuple.empty(), name="empty", grouping_strategy=GroupingStrategy.NONE)




import abc
from .io import super_open as super_open
from .print import TeeMultiOutput as TeeMultiOutput, debug as debug
from collections.abc import Callable as Callable
from contextlib import AbstractAsyncContextManager, AbstractContextManager
from typing import Any, IO, TextIO, TypeVar

T = TypeVar('T')

class AbstractBothContextManager[T](AbstractContextManager[T], AbstractAsyncContextManager[T], metaclass=abc.ABCMeta):
    """ Abstract base class for context managers that support both synchronous and asynchronous usage. """

class LogToFile(AbstractBothContextManager['LogToFile']):
    ''' Context manager to log to a file.

\tThis context manager allows you to temporarily log output to a file while still printing normally.
\tThe file will receive log messages without ANSI color codes.

\tArgs:
\t\tpath (str): Path to the log file
\t\tmode (str): Mode to open the file in (default: "w")
\t\tencoding (str): Encoding to use for the file (default: "utf-8")
\t\ttee_stdout (bool): Whether to redirect stdout to the file (default: True)
\t\ttee_stderr (bool): Whether to redirect stderr to the file (default: True)
\t\tignore_lineup (bool): Whether to ignore lines containing LINE_UP escape sequence in files (default: False)
\t\trestore_on_exit (bool): Whether to restore original stdout/stderr on exit (default: False)
\t\t\tThis ctx uses TeeMultiOutput which handles closed files gracefully, so restoring is not mandatory.

\tExamples:
\t\t.. code-block:: python

\t\t\t> import stouputils as stp
\t\t\t> with stp.LogToFile("output.log"):
\t\t\t>     stp.info("This will be logged to output.log and printed normally")
\t\t\t>     print("This will also be logged")

\t\t\t> with stp.LogToFile("output.log") as log_ctx:
\t\t\t>     stp.warning("This will be logged to output.log and printed normally")
\t\t\t>     log_ctx.change_file("new_file.log")
\t\t\t>     print("This will be logged to new_file.log")
\t'''
    path: str
    mode: str
    encoding: str
    tee_stdout: bool
    tee_stderr: bool
    ignore_lineup: bool
    restore_on_exit: bool
    file: IO[Any]
    original_stdout: TextIO
    original_stderr: TextIO
    def __init__(self, path: str, mode: str = 'w', encoding: str = 'utf-8', tee_stdout: bool = True, tee_stderr: bool = True, ignore_lineup: bool = True, restore_on_exit: bool = False) -> None: ...
    def __enter__(self) -> LogToFile:
        """ Enter context manager which opens the log file and redirects stdout/stderr """
    def __exit__(self, exc_type: type[BaseException] | None, exc_val: BaseException | None, exc_tb: Any | None) -> None:
        """ Exit context manager which closes the log file and restores stdout/stderr """
    async def __aenter__(self) -> LogToFile:
        """ Enter async context manager which opens the log file and redirects stdout/stderr """
    async def __aexit__(self, exc_type: type[BaseException] | None, exc_val: BaseException | None, exc_tb: Any | None) -> None:
        """ Exit async context manager which closes the log file and restores stdout/stderr """
    def change_file(self, new_path: str) -> None:
        """ Change the log file to a new path.

\t\tArgs:
\t\t\tnew_path (str): New path to the log file
\t\t"""
    @staticmethod
    def common(logs_folder: str, filepath: str, func: Callable[..., Any], *args: Any, **kwargs: Any) -> Any:
        ''' Common code used at the beginning of a program to launch main function

\t\tArgs:
\t\t\tlogs_folder (str): Folder to store logs in
\t\t\tfilepath    (str): Path to the main function
\t\t\tfunc        (Callable[..., Any]): Main function to launch
\t\t\t*args       (tuple[Any, ...]): Arguments to pass to the main function
\t\t\t**kwargs    (dict[str, Any]): Keyword arguments to pass to the main function
\t\tReturns:
\t\t\tAny: Return value of the main function

\t\tExamples:
\t\t\t>>> if __name__ == "__main__":
\t\t\t...     LogToFile.common(f"{ROOT}/logs", __file__, main)
\t\t'''

class MeasureTime(AbstractBothContextManager['MeasureTime']):
    ''' Context manager to measure execution time.

\tThis context manager measures the execution time of the code block it wraps
\tand prints the result using a specified print function.

\tArgs:
\t\tprint_func      (Callable): Function to use to print the execution time (e.g. debug, info, warning, error, etc.).
\t\tmessage         (str):      Message to display with the execution time. Defaults to "Execution time".
\t\tperf_counter    (bool):     Whether to use time.perf_counter_ns or time.time_ns. Defaults to True.

\tExamples:
\t\t.. code-block:: python

\t\t\t> import time
\t\t\t> import stouputils as stp
\t\t\t> with stp.MeasureTime(stp.info, message="My operation"):
\t\t\t...     time.sleep(0.5)
\t\t\t> # [INFO HH:MM:SS] My operation: 500.123ms (500123456ns)

\t\t\t> with stp.MeasureTime(): # Uses debug by default
\t\t\t...     time.sleep(0.1)
\t\t\t> # [DEBUG HH:MM:SS] Execution time: 100.456ms (100456789ns)
\t'''
    print_func: Callable[..., None]
    message: str
    perf_counter: bool
    ns: Callable[[], int]
    start_ns: int
    def __init__(self, print_func: Callable[..., None] = ..., message: str = 'Execution time', perf_counter: bool = True) -> None: ...
    def __enter__(self) -> MeasureTime:
        """ Enter context manager, record start time """
    def __exit__(self, exc_type: type[BaseException] | None, exc_val: BaseException | None, exc_tb: Any | None) -> None:
        """ Exit context manager, calculate duration and print """
    async def __aenter__(self) -> MeasureTime:
        """ Enter async context manager, record start time """
    async def __aexit__(self, exc_type: type[BaseException] | None, exc_val: BaseException | None, exc_tb: Any | None) -> None:
        """ Exit async context manager, calculate duration and print """

class Muffle(AbstractBothContextManager['Muffle']):
    ''' Context manager that temporarily silences output.
\t(No thread-safety guaranteed)

\tAlternative to stouputils.decorators.silent()

\tExamples:
\t\t>>> with Muffle():
\t\t...     print("This will not be printed")
\t'''
    mute_stderr: bool
    original_stdout: IO[Any]
    original_stderr: IO[Any]
    def __init__(self, mute_stderr: bool = False) -> None: ...
    def __enter__(self) -> Muffle:
        """ Enter context manager which redirects stdout and stderr to devnull """
    def __exit__(self, exc_type: type[BaseException] | None, exc_val: BaseException | None, exc_tb: Any | None) -> None:
        """ Exit context manager which restores original stdout and stderr """
    async def __aenter__(self) -> Muffle:
        """ Enter async context manager which redirects stdout and stderr to devnull """
    async def __aexit__(self, exc_type: type[BaseException] | None, exc_val: BaseException | None, exc_tb: Any | None) -> None:
        """ Exit async context manager which restores original stdout and stderr """

class DoNothing(AbstractBothContextManager['DoNothing']):
    ''' Context manager that does nothing.

\tThis is a no-op context manager that can be used as a placeholder
\tor for conditional context management.

\tDifferent from contextlib.nullcontext because it handles args and kwargs,
\talong with **async** context management.

\tExamples:
\t\t>>> with DoNothing():
\t\t...     print("This will be printed normally")
\t\tThis will be printed normally

\t\t>>> # Conditional context management
\t\t>>> some_condition = True
\t\t>>> ctx = DoNothing() if some_condition else Muffle()
\t\t>>> with ctx:
\t\t...     print("May or may not be printed depending on condition")
\t\tMay or may not be printed depending on condition
\t'''
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        """ No initialization needed, this is a no-op context manager """
    def __enter__(self) -> DoNothing:
        """ Enter context manager (does nothing) """
    def __exit__(self, exc_type: Any, exc_val: Any, exc_tb: Any) -> None:
        """ Exit context manager (does nothing) """
    async def __aenter__(self) -> DoNothing:
        """ Enter async context manager (does nothing) """
    async def __aexit__(self, *excinfo: Any) -> None:
        """ Exit async context manager (does nothing) """
NullContextManager = DoNothing

class SetMPStartMethod(AbstractBothContextManager['SetMPStartMethod']):
    ''' Context manager to temporarily set multiprocessing start method.

\tThis context manager allows you to temporarily change the multiprocessing start method
\tand automatically restores the original method when exiting the context.

\tArgs:
\t\tstart_method (str): The start method to use: "spawn", "fork", or "forkserver"

\tExamples:
\t\t.. code-block:: python

\t\t\t> import multiprocessing as mp
\t\t\t> import stouputils as stp
\t\t\t> # Temporarily use spawn method
\t\t\t> with stp.SetMPStartMethod("spawn"):
\t\t\t> ...     # Your multiprocessing code here
\t\t\t> ...     pass

\t\t\t> # Original method is automatically restored
\t'''
    start_method: str | None
    old_method: str | None
    def __init__(self, start_method: str | None) -> None: ...
    def __enter__(self) -> SetMPStartMethod:
        """ Enter context manager which sets the start method """
    def __exit__(self, exc_type: Any, exc_val: Any, exc_tb: Any) -> None:
        """ Exit context manager which restores the original start method """
    async def __aenter__(self) -> SetMPStartMethod:
        """ Enter async context manager which sets the start method """
    async def __aexit__(self, exc_type: Any, exc_val: Any, exc_tb: Any) -> None:
        """ Exit async context manager which restores the original start method """



from .print import CYAN as CYAN, GREEN as GREEN, RESET as RESET, YELLOW as YELLOW

def show_version(main_package: str = 'stouputils', primary_color: str = ..., secondary_color: str = ..., max_depth: int = 2) -> None:
    ''' Print the version of the main package and its dependencies.

\tUsed by the "stouputils --version" command.

\tArgs:
\t\tmain_package\t(str):\tName of the main package to show version for
\t\tprimary_color\t(str):\tColor to use for the primary package name
\t\tsecondary_color\t(str):\tColor to use for the secondary package names
\t\tmax_depth\t\t(int):\tMaximum depth for dependency tree (<= 2 for flat, >=3 for tree)
\t'''
def show_version_cli() -> None:
    ''' Handle the "stouputils --version" CLI command '''



"""
This module provides utilities for file management.

- get_root_path: Get the absolute path of the directory
- relative_path: Get the relative path of a file relative to a given directory
- json_dump: Writes the provided data to a JSON file with a specified indentation depth.
- json_load: Load a JSON file from the given path
- csv_dump: Writes data to a CSV file with customizable options
- csv_load: Load a CSV file from the given path
- super_copy: Copy a file (or a folder) from the source to the destination (always create the directory)
- super_open: Open a file with the given mode, creating the directory if it doesn't exist (only if writing)
- replace_tilde: Replace the "~" by the user's home directory
- clean_path: Clean the path by replacing backslashes with forward slashes and simplifying the path

.. image:: https://raw.githubusercontent.com/Stoupy51/stouputils/refs/heads/main/assets/io_module.gif
  :alt: stouputils io examples
"""

# Imports
import csv
import json
import os
import re
import shutil
from io import StringIO
from typing import IO, Any


# Function that takes a relative path and returns the absolute path of the directory
def get_root_path(relative_path: str, go_up: int = 0) -> str:
	""" Get the absolute path of the directory.
	Usually used to get the root path of the project using the __file__ variable.

	Args:
		relative_path   (str): The path to get the absolute directory path from
		go_up           (int): Number of parent directories to go up (default: 0)
	Returns:
		str: The absolute path of the directory

	Examples:

		.. code-block:: python

			> get_root_path(__file__)
			'C:/Users/Alexandre-PC/AppData/Local/Programs/Python/Python310/lib/site-packages/stouputils'

			> get_root_path(__file__, 3)
			'C:/Users/Alexandre-PC/AppData/Local/Programs/Python/Python310'
	"""
	return clean_path(
		os.path.dirname(os.path.abspath(relative_path))
		+ "/.." * go_up
	) or "."

# Function that returns the relative path of a file
def relative_path(file_path: str, relative_to: str = "") -> str:
	""" Get the relative path of a file relative to a given directory.

	Args:
		file_path     (str): The path to get the relative path from
		relative_to   (str): The path to get the relative path to (default: current working directory -> os.getcwd())
	Returns:
		str: The relative path of the file
	Examples:

		>>> relative_path("D:/some/random/path/stouputils/io.py", "D:\\\\some")
		'random/path/stouputils/io.py'
		>>> relative_path("D:/some/random/path/stouputils/io.py", "D:\\\\some\\\\")
		'random/path/stouputils/io.py'
	"""
	if not relative_to:
		relative_to = os.getcwd()
	file_path = clean_path(file_path)
	relative_to = clean_path(relative_to)
	if file_path.startswith(relative_to):
		return clean_path(os.path.relpath(file_path, relative_to)) or "."
	else:
		return file_path or "."

# JSON dump with indentation for levels
def json_dump(
	data: Any,
	file: IO[Any] | str | None = None,
	max_level: int | None = 2,
	indent: str | int = '\t',
	suffix: str = "\n",
	ensure_ascii: bool = False
) -> str:
	r""" Writes the provided data to a JSON file with a specified indentation depth.
	For instance, setting max_level to 2 will limit the indentation to 2 levels.

	Args:
		data		(Any): 				The data to dump (usually a dict or a list)
		file		(IO[Any] | str): 	The file object or path to dump the data to
		max_level	(int | None):		The depth of indentation to stop at (-1 for infinite), None will default to 2
		indent		(str | int):		The indentation character (default: '\t')
		suffix		(str):				The suffix to add at the end of the string (default: '\n')
		ensure_ascii (bool):			Whether to escape non-ASCII characters (default: False)
	Returns:
		str: The content of the file in every case

	>>> json_dump({"a": [[1,2,3]], "b": 2}, max_level = 0)
	'{"a": [[1,2,3]],"b": 2}\n'
	>>> json_dump({"a": [[1,2,3]], "b": 2}, max_level = 1)
	'{\n\t"a": [[1,2,3]],\n\t"b": 2\n}\n'
	>>> json_dump({"a": [[1,2,3]], "b": 2}, max_level = 2)
	'{\n\t"a": [\n\t\t[1,2,3]\n\t],\n\t"b": 2\n}\n'
	>>> json_dump({"a": [[1,2,3]], "b": 2}, max_level = 3)
	'{\n\t"a": [\n\t\t[\n\t\t\t1,\n\t\t\t2,\n\t\t\t3\n\t\t]\n\t],\n\t"b": 2\n}\n'
	>>> json_dump({"éà": "üñ"}, ensure_ascii = True, max_level = 0)
	'{"\\u00e9\\u00e0": "\\u00fc\\u00f1"}\n'
	>>> json_dump({"éà": "üñ"}, ensure_ascii = False, max_level = 0)
	'{"éà": "üñ"}\n'
	"""
	# Handle None values for max_level
	if max_level is None:
		max_level = 2

	# Dump content with 2-space indent and replace it with the desired indent
	content: str = json.dumps(data, indent=indent, ensure_ascii=ensure_ascii)

	# Limit max depth of indentation
	if max_level > -1:
		escape: str = re.escape(indent if isinstance(indent, str) else ' '*indent)
		pattern: re.Pattern[str] = re.compile(
			r"\n" + escape + "{" + str(max_level + 1) + r",}(.*)"
			r"|\n" + escape + "{" + str(max_level) + r"}([}\]])"
		)
		content = pattern.sub(r"\1\2", content)

	# Final newline and write
	content += suffix
	if file:
		if isinstance(file, str):
			with super_open(file, "w") as f:
				f.write(content)
		else:
			file.write(content)
	return content

# JSON load from file path
def json_load(file_path: str) -> Any:
	""" Load a JSON file from the given path

	Args:
		file_path (str): The path to the JSON file
	Returns:
		Any: The content of the JSON file
	"""
	with open(file_path) as f:
		return json.loads(f.read())

# CSV dump to file
def csv_dump(
	data: Any,
	file: IO[Any] | str | None = None,
	delimiter: str = ',',
	has_header: bool = True,
	index: bool = False,
	*args: Any,
	**kwargs: Any
) -> str:
	""" Writes data to a CSV file with customizable options and returns the CSV content as a string.

	Args:
		data		(list[list[Any]] | list[dict[str, Any]] | pd.DataFrame | pl.DataFrame):
						The data to write, either a list of lists, list of dicts, pandas DataFrame, or Polars DataFrame
		file		(IO[Any] | str): The file object or path to dump the data to
		delimiter	(str): The delimiter to use (default: ',')
		has_header	(bool): Whether to include headers (default: True, applies to dict and DataFrame data)
		index		(bool): Whether to include the index (default: False, only applies to pandas DataFrame)
		*args		(Any): Additional positional arguments to pass to the underlying CSV writer or DataFrame method
		**kwargs	(Any): Additional keyword arguments to pass to the underlying CSV writer or DataFrame method
	Returns:
		str: The CSV content as a string

	Examples:

		>>> csv_dump([["a", "b", "c"], [1, 2, 3], [4, 5, 6]])
		'a,b,c\\r\\n1,2,3\\r\\n4,5,6\\r\\n'

		>>> csv_dump([{"name": "Alice", "age": 30}, {"name": "Bob", "age": 25}])
		'name,age\\r\\nAlice,30\\r\\nBob,25\\r\\n'
	"""
	if isinstance(data, str | bytes | dict):
		raise ValueError("Data must be a list of lists, list of dicts, pandas DataFrame, or Polars DataFrame")
	output = StringIO()
	done: bool = False

	# Handle Polars DataFrame
	try:
		import polars as pl  # type: ignore
		if isinstance(data, pl.DataFrame):
			copy_kwargs = kwargs.copy()
			copy_kwargs.setdefault("separator", delimiter)
			copy_kwargs.setdefault("include_header", has_header)
			data.write_csv(output, *args, **copy_kwargs)
			done = True
	except Exception:
		pass

	# Handle pandas DataFrame
	if not done:
		try:
			import pandas as pd  # type: ignore
			if isinstance(data, pd.DataFrame):
				copy_kwargs = kwargs.copy()
				copy_kwargs.setdefault("index", index)
				copy_kwargs.setdefault("sep", delimiter)
				copy_kwargs.setdefault("header", has_header)
				data.to_csv(output, *args, **copy_kwargs)
		except Exception:
			pass

	if not done:
		# Handle list of dicts
		data = list(data)	# Ensure list and not other iterable
		if isinstance(data[0], dict):
			fieldnames = list(data[0].keys()) # type: ignore
			kwargs.setdefault("fieldnames", fieldnames)
			kwargs.setdefault("delimiter", delimiter)
			dict_writer = csv.DictWriter(output, *args, **kwargs)
			if has_header:
				dict_writer.writeheader()
			dict_writer.writerows(data)  # type: ignore
			done = True

		# Handle list of lists
		else:
			kwargs.setdefault("delimiter", delimiter)
			list_writer = csv.writer(output, *args, **kwargs)
			list_writer.writerows(data) # type: ignore
			done = True

	# If still not done, raise error
	if not done:
		output.close()
		raise ValueError(f"Data must be a list of lists, list of dicts, pandas DataFrame, or Polars DataFrame, got {type(data)} instead")

	# Get content and write to file if needed
	content: str = output.getvalue()
	if file:
		if isinstance(file, str):
			with super_open(file, "w") as f:
				f.write(content)
		else:
			file.write(content)
	output.close()
	return content

# CSV load from file path
def csv_load(file_path: str, delimiter: str = ',', has_header: bool = True, as_dict: bool = False, as_dataframe: bool = False, use_polars: bool = False, *args: Any, **kwargs: Any) -> Any:
	""" Load a CSV file from the given path

	Args:
		file_path (str): The path to the CSV file
		delimiter (str): The delimiter used in the CSV (default: ',')
		has_header (bool): Whether the CSV has a header row (default: True)
		as_dict (bool): Whether to return data as list of dicts (default: False)
		as_dataframe (bool): Whether to return data as a DataFrame (default: False)
		use_polars (bool): Whether to use Polars instead of pandas for DataFrame (default: False, requires polars)
		*args: Additional positional arguments to pass to the underlying CSV reader or DataFrame method
		**kwargs: Additional keyword arguments to pass to the underlying CSV reader or DataFrame method
	Returns:
		list[list[str]] | list[dict[str, str]] | pd.DataFrame | pl.DataFrame: The content of the CSV file

	Examples:

		.. code-block:: python

			> Assuming "test.csv" contains: a,b,c\\n1,2,3\\n4,5,6
			> csv_load("test.csv")
			[['1', '2', '3'], ['4', '5', '6']]

			> csv_load("test.csv", as_dict=True)
			[{'a': '1', 'b': '2', 'c': '3'}, {'a': '4', 'b': '5', 'c': '6'}]

			> csv_load("test.csv", as_dataframe=True)
			   a  b  c
			0  1  2  3
			1  4  5  6

		.. code-block:: console

			> csv_load("test.csv", as_dataframe=True, use_polars=True)
			shape: (2, 3)
			┌─────┬─────┬─────┐
			│ a   ┆ b   ┆ c   │
			│ --- ┆ --- ┆ --- │
			│ i64 ┆ i64 ┆ i64 │
			╞═════╪═════╪═════╡
			│ 1   ┆ 2   ┆ 3   │
			│ 4   ┆ 5   ┆ 6   │
			└─────┴─────┴─────┘
	"""  # noqa: E101
	# Handle DataFrame loading
	if as_dataframe:
		if use_polars:
			import polars as pl  # type: ignore
			if not os.path.exists(file_path):
				return pl.DataFrame() # type: ignore
			kwargs.setdefault("separator", delimiter)
			kwargs.setdefault("has_header", has_header)
			return pl.read_csv(file_path, *args, **kwargs) # type: ignore
		else:
			import pandas as pd  # type: ignore
			if not os.path.exists(file_path):
				return pd.DataFrame() # type: ignore
			kwargs.setdefault("sep", delimiter)
			kwargs.setdefault("header", 0 if has_header else None)
			return pd.read_csv(file_path, *args, **kwargs) # type: ignore

	# Handle dict or list
	if not os.path.exists(file_path):
		return []
	with super_open(file_path, "r") as f:
		if as_dict or has_header:
			kwargs.setdefault("delimiter", delimiter)
			reader = csv.DictReader(f, *args, **kwargs)
			return list(reader)
		else:
			kwargs.setdefault("delimiter", delimiter)
			reader = csv.reader(f, *args, **kwargs)
			return list(reader)

# For easy file copy
def super_copy(src: str, dst: str, create_dir: bool = True, symlink: bool = False) -> str:
	""" Copy a file (or a folder) from the source to the destination

	Args:
		src         (str):  The source path
		dst         (str):  The destination path
		create_dir  (bool): Whether to create the directory if it doesn't exist (default: True)
		symlink     (bool): Whether to create a symlink instead of copying (Linux only, default: True)
	Returns:
		str: The destination path
	"""
	# Disable symlink functionality on Windows as it uses shortcuts instead of proper symlinks
	if os.name == "nt":
		symlink = False

	# Create destination directory if needed
	if create_dir:
		os.makedirs(os.path.dirname(dst), exist_ok=True)

	# Handle directory copying
	if os.path.isdir(src):
		if symlink:

			# Remove existing destination if it's different from source
			if os.path.exists(dst):
				if os.path.samefile(src, dst) is False:
					if os.path.isdir(dst):
						shutil.rmtree(dst)
					else:
						os.remove(dst)
					return os.symlink(src.rstrip('/'), dst.rstrip('/'), target_is_directory=True) or dst
			else:
				return os.symlink(src.rstrip('/'), dst.rstrip('/'), target_is_directory=True) or dst

		# Regular directory copy
		else:
			return shutil.copytree(src, dst, dirs_exist_ok = True)

	# Handle file copying
	else:
		if symlink:

			# Remove existing destination if it's different from source
			if os.path.exists(dst):
				if os.path.samefile(src, dst) is False:
					os.remove(dst)
					return os.symlink(src, dst, target_is_directory=False) or dst
			else:
				return os.symlink(src, dst, target_is_directory=False) or dst

		# Regular file copy
		else:
			return shutil.copy(src, dst)
	return ""

# For easy file management
def super_open(file_path: str, mode: str, encoding: str = "utf-8") -> IO[Any]:
	""" Open a file with the given mode, creating the directory if it doesn't exist (only if writing)

	Args:
		file_path	(str): The path to the file
		mode		(str): The mode to open the file with, ex: "w", "r", "a", "wb", "rb", "ab"
		encoding	(str): The encoding to use when opening the file (default: "utf-8")
	Returns:
		open: The file object, ready to be used
	"""
	# Make directory
	file_path = clean_path(file_path)
	if "/" in file_path and ("w" in mode or "a" in mode):
		os.makedirs(os.path.dirname(file_path), exist_ok=True)

	# Open file and return
	if "b" in mode:
		return open(file_path, mode)
	else:
		return open(file_path, mode, encoding = encoding) # Always use utf-8 encoding to avoid issues

def read_file(file_path: str, encoding: str = "utf-8") -> str:
	""" Read the content of a file and return it as a string

	Args:
		file_path (str): The path to the file
		encoding  (str): The encoding to use when opening the file (default: "utf-8")
	Returns:
		str: The content of the file
	"""
	with super_open(file_path, "r", encoding=encoding) as f:
		return f.read()

# Function that replace the "~" by the user's home directory
def replace_tilde(path: str) -> str:
	""" Replace the "~" by the user's home directory

	Args:
		path (str): The path to replace the "~" by the user's home directory
	Returns:
		str: The path with the "~" replaced by the user's home directory
	Examples:

		.. code-block:: python

			> replace_tilde("~/Documents/test.txt")
			'/home/user/Documents/test.txt'
	"""
	return path.replace("~", os.path.expanduser("~")).replace("\\", "/")

# Utility function to clean the path
def clean_path(file_path: str, trailing_slash: bool = True) -> str:
	""" Clean the path by replacing backslashes with forward slashes and simplifying the path

	Args:
		file_path (str): The path to clean
		trailing_slash (bool): Whether to keep the trailing slash, ex: "test/" -> "test/"
	Returns:
		str: The cleaned path
	Examples:
		>>> clean_path("C:\\\\Users\\\\Stoupy\\\\Documents\\\\test.txt")
		'C:/Users/Stoupy/Documents/test.txt'

		>>> clean_path("Some Folder////")
		'Some Folder/'

		>>> clean_path("test/uwu/1/../../")
		'test/'

		>>> clean_path("some/./folder/../")
		'some/'

		>>> clean_path("folder1/folder2/../../folder3")
		'folder3'

		>>> clean_path("./test/./folder/")
		'test/folder/'

		>>> clean_path("C:/folder1\\\\folder2")
		'C:/folder1/folder2'
	"""
	# Replace tilde
	file_path = replace_tilde(str(file_path))

	# Check if original path ends with slash
	ends_with_slash: bool = file_path.endswith('/') or file_path.endswith('\\')

	# Use os.path.normpath to clean up the path
	file_path = os.path.normpath(file_path)

	# Convert backslashes to forward slashes
	file_path = file_path.replace(os.sep, '/')

	# Add trailing slash back if original had one
	if ends_with_slash and not file_path.endswith('/'):
		file_path += '/'

	# Remove trailing slash if requested
	if not trailing_slash and file_path.endswith('/'):
		file_path = file_path[:-1]

	# Return the cleaned path
	return file_path if file_path != "." else ""





# pyright: reportUnusedImport=false
# ruff: noqa: F401

# Imports
from .common import Any, NDArray, check_image, cv2, np


# Functions
def median_blur_image(image: NDArray[Any], kernel_size: int = 7, iterations: int = 1) -> NDArray[Any]:
	""" Apply median blur to an image.

	Args:
		image         (NDArray[Any]):  Image to apply median blur
		kernel_size   (int):         Kernel size for the median blur
		iterations    (int):         Number of iterations for the median blur
	Returns:
		NDArray[Any]: Image with median blur applied

	>>> ## Basic tests
	>>> image = np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]], dtype=np.uint8)
	>>> adjusted = median_blur_image(image, kernel_size=7, iterations=1)
	>>> adjusted.tolist()
	[[3, 3, 3], [4, 5, 6], [7, 7, 7]]
	>>> adjusted.shape == image.shape
	True
	>>> adjusted.dtype == image.dtype
	True

	>>> median_blur_image(image, kernel_size=3, iterations=1).tolist()
	[[2, 3, 3], [4, 5, 6], [7, 7, 8]]
	>>> median_blur_image(image, kernel_size=3, iterations=2).tolist()
	[[3, 3, 3], [4, 5, 6], [7, 7, 7]]
	>>> median_blur_image(image, kernel_size=3, iterations=5).tolist()
	[[3, 3, 3], [4, 5, 6], [7, 7, 7]]

	>>> ## Test invalid inputs
	>>> median_blur_image("not an image")
	Traceback (most recent call last):
		...
	AssertionError: Image must be a numpy array
	"""
	# Check input data
	check_image(image, ignore_dtype=True)

	# Apply median blur
	for _ in range(iterations):
		image = cv2.medianBlur(image, kernel_size)

	# Return the image
	return image




"""
This module provides utility functions for parallel processing, such as:

- multiprocessing(): Execute a function in parallel using multiprocessing
- multithreading(): Execute a function in parallel using multithreading
- run_in_subprocess(): Execute a function in a subprocess with args and kwargs

I highly encourage you to read the function docstrings to understand when to use each method.

.. image:: https://raw.githubusercontent.com/Stoupy51/stouputils/refs/heads/main/assets/parallel_module.gif
  :alt: stouputils parallel examples
"""

# Imports
import os
import time
from collections.abc import Callable, Iterable
from typing import Any, TypeVar, cast

from .ctx import SetMPStartMethod
from .print import BAR_FORMAT, MAGENTA


# Small test functions for doctests
def doctest_square(x: int) -> int:
	return x * x
def doctest_slow(x: int) -> int:
	time.sleep(0.1)
	return x

# Constants
CPU_COUNT: int = cast(int, os.cpu_count())
T = TypeVar("T")
R = TypeVar("R")

# Functions
def multiprocessing[T, R](
	func: Callable[..., R] | list[Callable[..., R]],
	args: Iterable[T],
	use_starmap: bool = False,
	chunksize: int = 1,
	desc: str = "",
	max_workers: int | float = CPU_COUNT,
	delay_first_calls: float = 0,
	color: str = MAGENTA,
	bar_format: str = BAR_FORMAT,
	ascii: bool = False,
) -> list[R]:
	r""" Method to execute a function in parallel using multiprocessing

	- For CPU-bound operations where the GIL (Global Interpreter Lock) is a bottleneck.
	- When the task can be divided into smaller, independent sub-tasks that can be executed concurrently.
	- For computationally intensive tasks like scientific simulations, data analysis, or machine learning workloads.

	Args:
		func				(Callable | list[Callable]):	Function to execute, or list of functions (one per argument)
		args				(Iterable):			Iterable of arguments to pass to the function(s)
		use_starmap			(bool):				Whether to use starmap or not (Defaults to False):
			True means the function will be called like func(\*args[i]) instead of func(args[i])
		chunksize			(int):				Number of arguments to process at a time
			(Defaults to 1 for proper progress bar display)
		desc				(str):				Description displayed in the progress bar
			(if not provided no progress bar will be displayed)
		max_workers			(int | float):		Number of workers to use (Defaults to CPU_COUNT), -1 means CPU_COUNT.
			If float between 0 and 1, it's treated as a percentage of CPU_COUNT.
			If negative float between -1 and 0, it's treated as a percentage of len(args).
		delay_first_calls	(float):			Apply i*delay_first_calls seconds delay to the first "max_workers" calls.
			For instance, the first process will be delayed by 0 seconds, the second by 1 second, etc.
			(Defaults to 0): This can be useful to avoid functions being called in the same second.
		color				(str):				Color of the progress bar (Defaults to MAGENTA)
		bar_format			(str):				Format of the progress bar (Defaults to BAR_FORMAT)
		ascii				(bool):				Whether to use ASCII or Unicode characters for the progress bar

	Returns:
		list[object]:	Results of the function execution

	Examples:
		.. code-block:: python

			> multiprocessing(doctest_square, args=[1, 2, 3])
			[1, 4, 9]

			> multiprocessing(int.__mul__, [(1,2), (3,4), (5,6)], use_starmap=True)
			[2, 12, 30]

			> # Using a list of functions (one per argument)
			> multiprocessing([doctest_square, doctest_square, doctest_square], [1, 2, 3])
			[1, 4, 9]

			> # Will process in parallel with progress bar
			> multiprocessing(doctest_slow, range(10), desc="Processing")
			[0, 1, 2, 3, 4, 5, 6, 7, 8, 9]

			> # Will process in parallel with progress bar and delay the first threads
			> multiprocessing(
			.     doctest_slow,
			.     range(10),
			.     desc="Processing with delay",
			.     max_workers=2,
			.     delay_first_calls=0.6
			. )
			[0, 1, 2, 3, 4, 5, 6, 7, 8, 9]
	"""
	# Imports
	import multiprocessing as mp
	from multiprocessing import Pool

	from tqdm.auto import tqdm
	from tqdm.contrib.concurrent import process_map  # pyright: ignore[reportUnknownVariableType]

	# Handle parameters
	args = list(args)  # Ensure we have a list (not other iterable)
	if max_workers == -1:
		max_workers = CPU_COUNT
	if isinstance(max_workers, float):
		if max_workers > 0:
			assert max_workers <= 1, "max_workers as positive float must be between 0 and 1 (percentage of CPU_COUNT)"
			max_workers = int(max_workers * CPU_COUNT)
		else:
			assert -1 <= max_workers < 0, "max_workers as negative float must be between -1 and 0 (percentage of len(args))"
			max_workers = int(-max_workers * len(args))
	verbose: bool = desc != ""
	desc, func, args = _handle_parameters(func, args, use_starmap, delay_first_calls, max_workers, desc, color)
	if bar_format == BAR_FORMAT:
		bar_format = bar_format.replace(MAGENTA, color)

	# Do multiprocessing only if there is more than 1 argument and more than 1 CPU
	if max_workers > 1 and len(args) > 1:
		def process() -> list[Any]:
			if verbose:
				return list(process_map(
					func, args, max_workers=max_workers, chunksize=chunksize, desc=desc, bar_format=bar_format, ascii=ascii
				)) # type: ignore
			else:
				with Pool(max_workers) as pool:
					return list(pool.map(func, args, chunksize=chunksize))	# type: ignore
		try:
			return process()
		except RuntimeError as e:
			if "SemLock created in a fork context is being shared with a process in a spawn context" in str(e):

				# Try with alternate start method
				with SetMPStartMethod("spawn" if mp.get_start_method() != "spawn" else "fork"):
					return process()
			else: # Re-raise if it's not the SemLock error
				raise

	# Single process execution
	else:
		if verbose:
			return [func(arg) for arg in tqdm(args, total=len(args), desc=desc, bar_format=bar_format, ascii=ascii)]
		else:
			return [func(arg) for arg in args]


def multithreading[T, R](
	func: Callable[..., R] | list[Callable[..., R]],
	args: Iterable[T],
	use_starmap: bool = False,
	desc: str = "",
	max_workers: int | float = CPU_COUNT,
	delay_first_calls: float = 0,
	color: str = MAGENTA,
	bar_format: str = BAR_FORMAT,
	ascii: bool = False,
	) -> list[R]:
	r""" Method to execute a function in parallel using multithreading, you should use it:

	- For I/O-bound operations where the GIL is not a bottleneck, such as network requests or disk operations.
	- When the task involves waiting for external resources, such as network responses or user input.
	- For operations that involve a lot of waiting, such as GUI event handling or handling user input.

	Args:
		func				(Callable | list[Callable]):	Function to execute, or list of functions (one per argument)
		args				(Iterable):			Iterable of arguments to pass to the function(s)
		use_starmap			(bool):				Whether to use starmap or not (Defaults to False):
			True means the function will be called like func(\*args[i]) instead of func(args[i])
		desc				(str):				Description displayed in the progress bar
			(if not provided no progress bar will be displayed)
		max_workers			(int | float):		Number of workers to use (Defaults to CPU_COUNT), -1 means CPU_COUNT.
			If float between 0 and 1, it's treated as a percentage of CPU_COUNT.
			If negative float between -1 and 0, it's treated as a percentage of len(args).
		delay_first_calls	(float):			Apply i*delay_first_calls seconds delay to the first "max_workers" calls.
			For instance with value to 1, the first thread will be delayed by 0 seconds, the second by 1 second, etc.
			(Defaults to 0): This can be useful to avoid functions being called in the same second.
		color				(str):				Color of the progress bar (Defaults to MAGENTA)
		bar_format			(str):				Format of the progress bar (Defaults to BAR_FORMAT)
		ascii				(bool):				Whether to use ASCII or Unicode characters for the progress bar

	Returns:
		list[object]:	Results of the function execution

	Examples:
		.. code-block:: python

			> multithreading(doctest_square, args=[1, 2, 3])
			[1, 4, 9]

			> multithreading(int.__mul__, [(1,2), (3,4), (5,6)], use_starmap=True)
			[2, 12, 30]

			> # Using a list of functions (one per argument)
			> multithreading([doctest_square, doctest_square, doctest_square], [1, 2, 3])
			[1, 4, 9]

			> # Will process in parallel with progress bar
			> multithreading(doctest_slow, range(10), desc="Threading")
			[0, 1, 2, 3, 4, 5, 6, 7, 8, 9]

			> # Will process in parallel with progress bar and delay the first threads
			> multithreading(
			.     doctest_slow,
			.     range(10),
			.     desc="Threading with delay",
			.     max_workers=2,
			.     delay_first_calls=0.6
			. )
			[0, 1, 2, 3, 4, 5, 6, 7, 8, 9]
	"""
	# Imports
	from concurrent.futures import ThreadPoolExecutor

	from tqdm.auto import tqdm

	# Handle parameters
	args = list(args)  # Ensure we have a list (not other iterable)
	if max_workers == -1:
		max_workers = CPU_COUNT
	if isinstance(max_workers, float):
		if max_workers > 0:
			assert max_workers <= 1, "max_workers as positive float must be between 0 and 1 (percentage of CPU_COUNT)"
			max_workers = int(max_workers * CPU_COUNT)
		else:
			assert -1 <= max_workers < 0, "max_workers as negative float must be between -1 and 0 (percentage of len(args))"
			max_workers = int(-max_workers * len(args))
	verbose: bool = desc != ""
	desc, func, args = _handle_parameters(func, args, use_starmap, delay_first_calls, max_workers, desc, color)
	if bar_format == BAR_FORMAT:
		bar_format = bar_format.replace(MAGENTA, color)

	# Do multithreading only if there is more than 1 argument and more than 1 CPU
	if max_workers > 1 and len(args) > 1:
		if verbose:
			with ThreadPoolExecutor(max_workers) as executor:
				return list(tqdm(executor.map(func, args), total=len(args), desc=desc, bar_format=bar_format, ascii=ascii))
		else:
			with ThreadPoolExecutor(max_workers) as executor:
				return list(executor.map(func, args))

	# Single process execution
	else:
		if verbose:
			return [func(arg) for arg in tqdm(args, total=len(args), desc=desc, bar_format=bar_format, ascii=ascii)]
		else:
			return [func(arg) for arg in args]


def run_in_subprocess[R](
	func: Callable[..., R],
	*args: Any,
	timeout: float | None = None,
	**kwargs: Any
) -> R:
	""" Execute a function in a subprocess with positional and keyword arguments.

	This is useful when you need to run a function in isolation to avoid memory leaks,
	resource conflicts, or to ensure a clean execution environment. The subprocess will
	be created, run the function with the provided arguments, and return the result.

	Args:
		func         (Callable):     The function to execute in a subprocess.
			(SHOULD BE A TOP-LEVEL FUNCTION TO BE PICKLABLE)
		*args        (Any):          Positional arguments to pass to the function.
		timeout      (float | None): Maximum time in seconds to wait for the subprocess.
			If None, wait indefinitely. If the subprocess exceeds this time, it will be terminated.
		**kwargs     (Any):          Keyword arguments to pass to the function.

	Returns:
		R: The return value of the function.

	Raises:
		RuntimeError: If the subprocess exits with a non-zero exit code or times out.
		TimeoutError: If the subprocess exceeds the specified timeout.

	Examples:
		.. code-block:: python

			> # Simple function execution
			> run_in_subprocess(doctest_square, 5)
			25

			> # Function with multiple arguments
			> def add(a: int, b: int) -> int:
			.     return a + b
			> run_in_subprocess(add, 10, 20)
			30

			> # Function with keyword arguments
			> def greet(name: str, greeting: str = "Hello") -> str:
			.     return f"{greeting}, {name}!"
			> run_in_subprocess(greet, "World", greeting="Hi")
			'Hi, World!'

			> # With timeout to prevent hanging
			> run_in_subprocess(some_gpu_func, data, timeout=300.0)
	"""
	import multiprocessing as mp
	from multiprocessing import Queue

	# Create a queue to get the result from the subprocess
	result_queue: Queue[R | Exception] = Queue()

	# Create and start the subprocess using the module-level wrapper
	process: mp.Process = mp.Process(
		target=_subprocess_wrapper,
		args=(result_queue, func, args, kwargs)
	)
	process.start()

	# Join with timeout to prevent indefinite hanging
	process.join(timeout=timeout)

	# Check if process is still alive (timed out)
	if process.is_alive():
		process.terminate()
		time.sleep(0.5)  # Give it a moment to terminate gracefully
		if process.is_alive():
			process.kill()
		process.join()
		raise TimeoutError(f"Subprocess exceeded timeout of {timeout} seconds and was terminated")

	# Check exit code
	if process.exitcode != 0:
		# Try to get any exception from the queue (non-blocking)
		error_msg = f"Subprocess failed with exit code {process.exitcode}"
		try:
			if not result_queue.empty():
				result_or_exception = result_queue.get_nowait()
				if isinstance(result_or_exception, Exception):
					raise result_or_exception
		except Exception:
			pass
		raise RuntimeError(error_msg)

	# Retrieve the result
	try:
		result_or_exception = result_queue.get_nowait()
		if isinstance(result_or_exception, Exception):
			raise result_or_exception
		return result_or_exception
	except Exception as e:
		raise RuntimeError("Subprocess did not return any result") from e


# "Private" function for subprocess wrapper (must be at module level for pickling on Windows)
def _subprocess_wrapper[R](
	result_queue: Any,
	func: Callable[..., R],
	args: tuple[Any, ...],
	kwargs: dict[str, Any]
) -> None:
	""" Wrapper function to execute the target function and store the result in the queue.

	Must be at module level to be pickable on Windows (spawn context).

	Args:
		result_queue (multiprocessing.Queue):  Queue to store the result or exception.
		func         (Callable):               The target function to execute.
		args         (tuple):                  Positional arguments for the function.
		kwargs       (dict):                   Keyword arguments for the function.
	"""
	try:
		result: R = func(*args, **kwargs)
		result_queue.put(result)
	except Exception as e:
		result_queue.put(e)

# "Private" function to use starmap
def _starmap[T, R](args: tuple[Callable[[T], R], list[T]]) -> R:
	r""" Private function to use starmap using args[0](\*args[1])

	Args:
		args (tuple): Tuple containing the function and the arguments list to pass to the function
	Returns:
		object: Result of the function execution
	"""
	func, arguments = args
	return func(*arguments)

# "Private" function to apply delay before calling the target function
def _delayed_call[T, R](args: tuple[Callable[[T], R], float, T]) -> R:
	""" Private function to apply delay before calling the target function

	Args:
		args (tuple): Tuple containing the function, delay in seconds, and the argument to pass to the function
	Returns:
		object: Result of the function execution
	"""
	func, delay, arg = args
	time.sleep(delay)
	return func(arg)

# "Private" function to handle parameters for multiprocessing or multithreading functions
def _handle_parameters[T, R](
	func: Callable[[T], R] | list[Callable[[T], R]],
	args: list[T],
	use_starmap: bool,
	delay_first_calls: float,
	max_workers: int,
	desc: str,
	color: str
) -> tuple[str, Callable[[T], R], list[T]]:
	r""" Private function to handle the parameters for multiprocessing or multithreading functions

	Args:
		func				(Callable | list[Callable]):	Function to execute, or list of functions (one per argument)
		args				(list):				List of arguments to pass to the function(s)
		use_starmap			(bool):				Whether to use starmap or not (Defaults to False):
			True means the function will be called like func(\*args[i]) instead of func(args[i])
		delay_first_calls	(int):				Apply i*delay_first_calls seconds delay to the first "max_workers" calls.
			For instance, the first process will be delayed by 0 seconds, the second by 1 second, etc. (Defaults to 0):
			This can be useful to avoid functions being called in the same second.
		max_workers			(int):				Number of workers to use (Defaults to CPU_COUNT)
		desc				(str):				Description of the function execution displayed in the progress bar
		color				(str):				Color of the progress bar

	Returns:
		tuple[str, Callable[[T], R], list[T]]:	Tuple containing the description, function, and arguments
	"""
	desc = color + desc

	# Handle list of functions: validate and convert to starmap format
	if isinstance(func, list):
		func = cast(list[Callable[[T], R]], func)
		assert len(func) == len(args), f"Length mismatch: {len(func)} functions but {len(args)} arguments"
		args = [(f, arg if use_starmap else (arg,)) for f, arg in zip(func, args, strict=False)] # type: ignore
		func = _starmap # type: ignore

	# If use_starmap is True, we use the _starmap function
	elif use_starmap:
		args = [(func, arg) for arg in args] # type: ignore
		func = _starmap # type: ignore

	# Prepare delayed function calls if delay_first_calls is set
	if delay_first_calls > 0:
		args = [
			(func, i * delay_first_calls if i < max_workers else 0, arg) # type: ignore
			for i, arg in enumerate(args)
		]
		func = _delayed_call  # type: ignore

	return desc, func, args # type: ignore




""" Deprecated functions and classes.

This module contains deprecated functions that have been replaced by new implementations
These functions are retained for backward compatibility and will log deprecation warnings when used.
"""

# Imports
from typing import Any

from .decorators import LogLevels, deprecated
from .io import csv_dump, csv_load, json_dump, json_load


# Deprecated functions
@deprecated(message="super_csv_dump has been renamed to csv_dump.", version="v1.8.0", error_log=LogLevels.WARNING)
def super_csv_dump(*args: Any, **kwargs: Any) -> Any:
	""" Deprecated function, use "csv_dump" instead. """
	return csv_dump(*args, **kwargs)


@deprecated(message="super_csv_load has been renamed to csv_load.", version="v1.8.0", error_log=LogLevels.WARNING)
def super_csv_load(*args: Any, **kwargs: Any) -> Any:
	""" Deprecated function, use "csv_load" instead. """
	return csv_load(*args, **kwargs)


@deprecated(message="super_json_dump has been renamed to json_dump.", version="v1.8.0", error_log=LogLevels.WARNING)
def super_json_dump(*args: Any, **kwargs: Any) -> Any:
	""" Deprecated function, use "json_dump" instead. """
	return json_dump(*args, **kwargs)


@deprecated(message="super_json_load has been renamed to json_load.", version="v1.8.0", error_log=LogLevels.WARNING)
def super_json_load(*args: Any, **kwargs: Any) -> Any:
	""" Deprecated function, use "json_load" instead. """
	return json_load(*args, **kwargs)





# pyright: reportUnknownMemberType=false
# pyright: reportUnknownVariableType=false
# pyright: reportUnknownArgumentType=false

# Imports
from typing import Literal

from .common import Any, NDArray, check_image, cv2, np
from ....ctx import Muffle


# Functions
def nlm_denoise_image(
	image: NDArray[Any],
	h: float = 10,
	template_window_size: int = 7,
	search_window_size: int = 21,
	ignore_dtype: bool = False
) -> NDArray[Any]:
	""" Apply Non-Local Means denoising to an image.

	This algorithm replaces each pixel with an average of similar pixels
	found anywhere in the image. It is highly effective for removing Gaussian noise
	while preserving edges and details.

	Args:
		image                  (NDArray[Any]):  Image to denoise
		h                      (float):       Filter strength (higher values remove more noise but may blur details)
		template_window_size   (int):         Size of the template window for patch comparison (should be odd)
		search_window_size     (int):         Size of the search window (should be odd)
		ignore_dtype           (bool):        Ignore the dtype check
	Returns:
		NDArray[Any]: Denoised image

	>>> ## Basic tests
	>>> image = np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]])
	>>> denoised = nlm_denoise_image(image.astype(np.uint8), 10, 3, 5)
	>>> denoised.shape == image.shape
	True

	>>> ## Test with colored image
	>>> rgb = np.random.randint(0, 256, (10, 10, 3), dtype=np.uint8)
	>>> denoised_rgb = nlm_denoise_image(rgb, 10, 5, 11)
	>>> denoised_rgb.shape == rgb.shape
	True

	>>> ## Test invalid inputs
	>>> nlm_denoise_image("not an image", 10)
	Traceback (most recent call last):
		...
	AssertionError: Image must be a numpy array

	>>> nlm_denoise_image(image.astype(np.uint8), "10")
	Traceback (most recent call last):
		...
	AssertionError: h must be a number, got <class 'str'>

	>>> nlm_denoise_image(image.astype(np.uint8), 10, 4)
	Traceback (most recent call last):
		...
	AssertionError: template_window_size must be odd, got 4
	"""
	# Check input data
	check_image(image, ignore_dtype=ignore_dtype)
	assert isinstance(h, float | int), f"h must be a number, got {type(h)}"
	assert template_window_size % 2 == 1, f"template_window_size must be odd, got {template_window_size}"
	assert search_window_size % 2 == 1, f"search_window_size must be odd, got {search_window_size}"

	# Apply Non-Local Means denoising based on image type
	if len(image.shape) == 2 or image.shape[2] == 1:  # Grayscale
		return cv2.fastNlMeansDenoising(
			image, None, float(h), template_window_size, search_window_size
		)
	else:  # Color
		return cv2.fastNlMeansDenoisingColored(
			image, None, float(h), float(h), template_window_size, search_window_size
		)


def bilateral_denoise_image(
	image: NDArray[Any],
	d: int = 9,
	sigma_color: float = 75,
	sigma_space: float = 75,
	ignore_dtype: bool = False
) -> NDArray[Any]:
	""" Apply Bilateral Filter denoising to an image.

	Bilateral filtering smooths images while preserving edges by considering
	both spatial proximity and color similarity between pixels.

	Args:
		image         (NDArray[Any]):  Image to denoise
		d             (int):         Diameter of each pixel neighborhood
		sigma_color   (float):       Filter sigma in the color space
		sigma_space   (float):       Filter sigma in the coordinate space
		ignore_dtype  (bool):        Ignore the dtype check
	Returns:
		NDArray[Any]: Denoised image

	>>> ## Basic tests
	>>> image = np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]])
	>>> denoised = bilateral_denoise_image(image.astype(np.uint8))
	>>> denoised.shape == image.shape
	True

	>>> ## Test with colored image
	>>> rgb = np.random.randint(0, 256, (10, 10, 3), dtype=np.uint8)
	>>> denoised_rgb = bilateral_denoise_image(rgb)
	>>> denoised_rgb.shape == rgb.shape
	True

	>>> ## Test invalid inputs
	>>> bilateral_denoise_image("not an image")
	Traceback (most recent call last):
		...
	AssertionError: Image must be a numpy array

	>>> bilateral_denoise_image(image.astype(np.uint8), "9")
	Traceback (most recent call last):
		...
	AssertionError: d must be a number, got <class 'str'>
	"""
	# Check input data
	check_image(image, ignore_dtype=ignore_dtype)
	assert isinstance(d, int), f"d must be a number, got {type(d)}"
	assert isinstance(sigma_color, float | int), f"sigma_color must be a number, got {type(sigma_color)}"
	assert isinstance(sigma_space, float | int), f"sigma_space must be a number, got {type(sigma_space)}"

	# Apply bilateral filter
	return cv2.bilateralFilter(image, d, sigma_color, sigma_space)


def tv_denoise_image(
	image: NDArray[Any],
	weight: float = 0.1,
	iterations: int = 30,
	method: Literal["chambolle", "bregman"] = "chambolle",
	ignore_dtype: bool = False
) -> NDArray[Any]:
	""" Apply Total Variation denoising to an image.

	Total Variation denoising removes noise while preserving sharp edges by
	minimizing the total variation of the image.

	Args:
		image         (NDArray[Any]):  Image to denoise
		weight        (float):       Denoising weight (higher values remove more noise)
		iterations    (int):         Number of iterations
		method        (str):         Method to use ("chambolle" or "bregman")
		ignore_dtype  (bool):        Ignore the dtype check
	Returns:
		NDArray[Any]: Denoised image

	>>> ## Basic tests
	>>> image = np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]])
	>>> denoised = tv_denoise_image(image.astype(np.uint8), 0.1, 30)
	>>> denoised.shape == image.shape
	True

	>>> ## Test invalid inputs
	>>> tv_denoise_image("not an image")
	Traceback (most recent call last):
		...
	AssertionError: Image must be a numpy array

	>>> tv_denoise_image(image.astype(np.uint8), "0.1")
	Traceback (most recent call last):
		...
	AssertionError: weight must be a number, got <class 'str'>
	"""
	# Check input data
	check_image(image, ignore_dtype=ignore_dtype)
	assert isinstance(weight, float | int), f"weight must be a number, got {type(weight)}"
	assert isinstance(iterations, int), f"iterations must be an integer, got {type(iterations)}"
	assert method in ["chambolle", "bregman"], f"method must be 'chambolle' or 'bregman', got {method}"

	# Import skimage for TV denoising
	try:
		from skimage.restoration import denoise_tv_bregman, denoise_tv_chambolle
	except ImportError as e:
		raise ImportError("scikit-image is required for TV denoising. Install with 'pip install scikit-image'") from e

	# Normalize image to [0, 1] for skimage functions
	is_int_type = np.issubdtype(image.dtype, np.integer)
	if is_int_type:
		img_norm = image.astype(np.float32) / 255.0
	else:
		img_norm = image.astype(np.float32)

	# Apply TV denoising based on method
	if method == "chambolle":
		denoised = denoise_tv_chambolle(
			img_norm,
			weight=weight,
			max_num_iter=iterations,
			channel_axis=-1 if len(image.shape) > 2 else None
		)
	else:
		denoised = denoise_tv_bregman(
			img_norm,
			weight=weight,
			max_num_iter=iterations,
			channel_axis=-1 if len(image.shape) > 2 else None
		)

	# Convert back to original data type
	if is_int_type:
		denoised = np.clip(denoised * 255, 0, 255).astype(image.dtype)
	else:
		denoised = denoised.astype(image.dtype)

	return denoised


def wavelet_denoise_image(
	image: NDArray[Any],
	sigma: float | None = None,
	wavelet: str = 'db1',
	mode: str = 'soft',
	wavelet_levels: int = 3,
	ignore_dtype: bool = False
) -> NDArray[Any]:
	""" Apply Wavelet denoising to an image.

	Wavelet denoising decomposes the image into wavelet coefficients,
	applies thresholding, and reconstructs the image with reduced noise.

	Args:
		image          (NDArray[Any]):  Image to denoise
		sigma          (float):       Noise standard deviation. If None, it's estimated from the image.
		wavelet        (str):         Wavelet to use
		mode           (str):         Thresholding mode ('soft' or 'hard')
		wavelet_levels (int):         Number of wavelet decomposition levels
		ignore_dtype   (bool):        Ignore the dtype check
	Returns:
		NDArray[Any]: Denoised image

	>>> ## Basic tests
	>>> import importlib.util
	>>> has_pywt = importlib.util.find_spec('pywt') is not None
	>>> image = np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]])
	>>> if has_pywt:
	...     denoised = wavelet_denoise_image(image.astype(np.uint8))
	...     denoised.shape == image.shape
	... else:
	...     True
	True

	>>> ## Test invalid inputs
	>>> wavelet_denoise_image("not an image")
	Traceback (most recent call last):
		...
	AssertionError: Image must be a numpy array

	>>> wavelet_denoise_image(image.astype(np.uint8), wavelet=123)
	Traceback (most recent call last):
		...
	AssertionError: wavelet must be a string, got <class 'int'>
	"""
	# Check input data
	check_image(image, ignore_dtype=ignore_dtype)
	if sigma is not None:
		assert isinstance(sigma, float | int), f"sigma must be a number or None, got {type(sigma)}"
	assert isinstance(wavelet, str), f"wavelet must be a string, got {type(wavelet)}"
	assert mode in ["soft", "hard"], f"mode must be 'soft' or 'hard', got {mode}"
	assert isinstance(wavelet_levels, int), f"wavelet_levels must be an integer, got {type(wavelet_levels)}"

	# Import skimage for wavelet denoising
	try:
		from skimage.restoration import denoise_wavelet

		# Check for PyWavelets dependency specifically
		try:
			import pywt  # type: ignore
		except ImportError as e:
			raise ImportError(
				"PyWavelets (pywt) is required for wavelet denoising. Install with 'pip install PyWavelets'", name="pywt"
			) from e
	except ImportError as e:
		if e.name != "pywt":
			raise ImportError("skimage is required for wavelet denoising. Install with 'pip install scikit-image'") from e
		else:
			raise e

	# Normalize image to [0, 1] for skimage functions
	is_int_type = np.issubdtype(image.dtype, np.integer)
	if is_int_type:
		img_norm = image.astype(np.float32) / 255.0
	else:
		img_norm = image.astype(np.float32)

	# Apply wavelet denoising
	with Muffle(mute_stderr=True):
		denoised = denoise_wavelet(
			img_norm, sigma=sigma, wavelet=wavelet, mode=mode,
			wavelet_levels=wavelet_levels, channel_axis=-1 if len(image.shape) > 2 else None
		)

	# Convert back to original data type
	if is_int_type:
		denoised = np.clip(denoised * 255, 0, 255).astype(image.dtype)
	else:
		denoised = denoised.astype(image.dtype)

	return denoised


def adaptive_denoise_image(
	image: NDArray[Any],
	method: Literal["nlm", "bilateral", "tv", "wavelet"] | str = "nlm",
	strength: float = 0.5,
	ignore_dtype: bool = False
) -> NDArray[Any]:
	""" Apply adaptive denoising to an image using the specified method.

	This is a convenience function that selects the appropriate denoising method
	and parameters based on the image content and noise level.

	Args:
		image         (NDArray[Any]):  Image to denoise
		method        (str):         Denoising method to use ("nlm", "bilateral", "tv", or "wavelet")
		strength      (float):       Denoising strength (0.0 to 1.0)
		ignore_dtype  (bool):        Ignore the dtype check
	Returns:
		NDArray[Any]: Denoised image

	>>> ## Basic tests
	>>> image = np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]])
	>>> denoised = adaptive_denoise_image(image.astype(np.uint8), "nlm", 0.5)
	>>> denoised.shape == image.shape
	True

	>>> ## Test invalid inputs
	>>> adaptive_denoise_image("not an image")
	Traceback (most recent call last):
		...
	AssertionError: Image must be a numpy array

	>>> adaptive_denoise_image(image.astype(np.uint8), "invalid_method")
	Traceback (most recent call last):
		...
	AssertionError: method must be one of: nlm, bilateral, tv, wavelet
	"""
	# Check input data
	check_image(image, ignore_dtype=ignore_dtype)
	valid_methods = ["nlm", "bilateral", "tv", "wavelet"]
	assert method in valid_methods, f"method must be one of: {', '.join(valid_methods)}"
	assert isinstance(strength, float | int), f"strength must be a number, got {type(strength)}"
	assert 0 <= strength <= 1, f"strength must be between 0 and 1, got {strength}"

	# Scale parameters based on strength
	if method == "bilateral":
		# sigma parameters scale from 30 (minimal) to 150 (strong)
		sigma = 30 + strength * 120
		return bilateral_denoise_image(
			image, d=9, sigma_color=sigma, sigma_space=sigma, ignore_dtype=ignore_dtype
		)

	elif method == "tv":
		# weight scales from 0.05 (minimal) to 0.5 (strong)
		weight = 0.05 + strength * 0.45
		return tv_denoise_image(
			image, weight=weight, iterations=30, ignore_dtype=ignore_dtype
		)

	elif method == "wavelet":
		# We'll estimate sigma from the image, but scale wavelet levels
		wavelet_levels = max(2, min(5, int(2 + strength * 3)))
		return wavelet_denoise_image(
			image, wavelet_levels=wavelet_levels, ignore_dtype=ignore_dtype
		)

	else:
		# h parameter scales from 5 (minimal) to 20 (strong)
		h = 5 + strength * 15
		return nlm_denoise_image(image, h=h, ignore_dtype=ignore_dtype)



"""
This module provides context managers for various utilities such as logging to a file,
measuring execution time, silencing output, and setting multiprocessing start methods.

- LogToFile: Context manager to log to a file every print call (with LINE_UP handling)
- MeasureTime: Context manager to measure execution time of a code block
- Muffle: Context manager that temporarily silences output (alternative to stouputils.decorators.silent())
- DoNothing: Context manager that does nothing (no-op)
- SetMPStartMethod: Context manager to temporarily set multiprocessing start method

.. image:: https://raw.githubusercontent.com/Stoupy51/stouputils/refs/heads/main/assets/ctx_module.gif
  :alt: stouputils ctx examples
"""

# Imports
from __future__ import annotations

import os
import sys
import time
from collections.abc import Callable
from contextlib import AbstractAsyncContextManager, AbstractContextManager
from typing import IO, Any, TextIO, TypeVar

from .io import super_open
from .print import TeeMultiOutput, debug

# Type variable for context managers
T = TypeVar("T")

# Abstract base class for context managers supporting both sync and async usage
class AbstractBothContextManager[T](AbstractContextManager[T], AbstractAsyncContextManager[T]):
    """ Abstract base class for context managers that support both synchronous and asynchronous usage. """
    pass

# Context manager to log to a file
class LogToFile(AbstractBothContextManager["LogToFile"]):
	""" Context manager to log to a file.

	This context manager allows you to temporarily log output to a file while still printing normally.
	The file will receive log messages without ANSI color codes.

	Args:
		path (str): Path to the log file
		mode (str): Mode to open the file in (default: "w")
		encoding (str): Encoding to use for the file (default: "utf-8")
		tee_stdout (bool): Whether to redirect stdout to the file (default: True)
		tee_stderr (bool): Whether to redirect stderr to the file (default: True)
		ignore_lineup (bool): Whether to ignore lines containing LINE_UP escape sequence in files (default: False)
		restore_on_exit (bool): Whether to restore original stdout/stderr on exit (default: False)
			This ctx uses TeeMultiOutput which handles closed files gracefully, so restoring is not mandatory.

	Examples:
		.. code-block:: python

			> import stouputils as stp
			> with stp.LogToFile("output.log"):
			>     stp.info("This will be logged to output.log and printed normally")
			>     print("This will also be logged")

			> with stp.LogToFile("output.log") as log_ctx:
			>     stp.warning("This will be logged to output.log and printed normally")
			>     log_ctx.change_file("new_file.log")
			>     print("This will be logged to new_file.log")
	"""
	def __init__(
		self,
		path: str,
		mode: str = "w",
		encoding: str = "utf-8",
		tee_stdout: bool = True,
		tee_stderr: bool = True,
		ignore_lineup: bool = True,
		restore_on_exit: bool = False
	) -> None:
		self.path: str = path
		""" Attribute remembering path to the log file """
		self.mode: str = mode
		""" Attribute remembering mode to open the file in """
		self.encoding: str = encoding
		""" Attribute remembering encoding to use for the file """
		self.tee_stdout: bool = tee_stdout
		""" Whether to redirect stdout to the file """
		self.tee_stderr: bool = tee_stderr
		""" Whether to redirect stderr to the file """
		self.ignore_lineup: bool = ignore_lineup
		""" Whether to ignore lines containing LINE_UP escape sequence in files """
		self.restore_on_exit: bool = restore_on_exit
		""" Whether to restore original stdout/stderr on exit.
		This ctx uses TeeMultiOutput which handles closed files gracefully, so restoring is not mandatory. """
		self.file: IO[Any]
		""" Attribute remembering opened file """
		self.original_stdout: TextIO
		""" Original stdout before redirection """
		self.original_stderr: TextIO
		""" Original stderr before redirection """

	def __enter__(self) -> LogToFile:
		""" Enter context manager which opens the log file and redirects stdout/stderr """
		# Open file
		self.file = super_open(self.path, mode=self.mode, encoding=self.encoding)

		# Redirect stdout and stderr if requested
		if self.tee_stdout:
			self.original_stdout = sys.stdout
			sys.stdout = TeeMultiOutput(self.original_stdout, self.file, ignore_lineup=self.ignore_lineup)
		if self.tee_stderr:
			self.original_stderr = sys.stderr
			sys.stderr = TeeMultiOutput(self.original_stderr, self.file, ignore_lineup=self.ignore_lineup)

		# Return self
		return self

	def __exit__(self, exc_type: type[BaseException]|None, exc_val: BaseException|None, exc_tb: Any|None) -> None:
		""" Exit context manager which closes the log file and restores stdout/stderr """
		# Restore original stdout and stderr (if requested)
		if self.restore_on_exit:
			if self.tee_stdout:
				sys.stdout = self.original_stdout
			if self.tee_stderr:
				sys.stderr = self.original_stderr

		# Close file
		self.file.close()

	async def __aenter__(self) -> LogToFile:
		""" Enter async context manager which opens the log file and redirects stdout/stderr """
		return self.__enter__()

	async def __aexit__(self, exc_type: type[BaseException]|None, exc_val: BaseException|None, exc_tb: Any|None) -> None:
		""" Exit async context manager which closes the log file and restores stdout/stderr """
		self.__exit__(exc_type, exc_val, exc_tb)

	def change_file(self, new_path: str) -> None:
		""" Change the log file to a new path.

		Args:
			new_path (str): New path to the log file
		"""
		# Close current file, open new file and redirect outputs
		self.file.close()
		self.path = new_path
		self.__enter__()

	@staticmethod
	def common(logs_folder: str, filepath: str, func: Callable[..., Any], *args: Any, **kwargs: Any) -> Any:
		""" Common code used at the beginning of a program to launch main function

		Args:
			logs_folder (str): Folder to store logs in
			filepath    (str): Path to the main function
			func        (Callable[..., Any]): Main function to launch
			*args       (tuple[Any, ...]): Arguments to pass to the main function
			**kwargs    (dict[str, Any]): Keyword arguments to pass to the main function
		Returns:
			Any: Return value of the main function

		Examples:
			>>> if __name__ == "__main__":
			...     LogToFile.common(f"{ROOT}/logs", __file__, main)
		"""
		# Import datetime
		from datetime import datetime

		# Build log file path
		file_basename: str = os.path.splitext(os.path.basename(filepath))[0]
		date_time: str = datetime.now().strftime("%Y-%m-%d_%H-%M-%S")
		date_str, time_str = date_time.split("_")
		log_filepath: str = f"{logs_folder}/{file_basename}/{date_str}/{time_str}.log"

		# Launch function with arguments if any
		with LogToFile(log_filepath):
			return func(*args, **kwargs)

# Context manager to measure execution time
class MeasureTime(AbstractBothContextManager["MeasureTime"]):
	""" Context manager to measure execution time.

	This context manager measures the execution time of the code block it wraps
	and prints the result using a specified print function.

	Args:
		print_func      (Callable): Function to use to print the execution time (e.g. debug, info, warning, error, etc.).
		message         (str):      Message to display with the execution time. Defaults to "Execution time".
		perf_counter    (bool):     Whether to use time.perf_counter_ns or time.time_ns. Defaults to True.

	Examples:
		.. code-block:: python

			> import time
			> import stouputils as stp
			> with stp.MeasureTime(stp.info, message="My operation"):
			...     time.sleep(0.5)
			> # [INFO HH:MM:SS] My operation: 500.123ms (500123456ns)

			> with stp.MeasureTime(): # Uses debug by default
			...     time.sleep(0.1)
			> # [DEBUG HH:MM:SS] Execution time: 100.456ms (100456789ns)
	"""
	def __init__(
		self,
		print_func: Callable[..., None] = debug,
		message: str = "Execution time",
		perf_counter: bool = True
	) -> None:
		self.print_func: Callable[..., None] = print_func
		""" Function to use for printing the execution time """
		self.message: str = message
		""" Message to display with the execution time """
		self.perf_counter: bool = perf_counter
		""" Whether to use time.perf_counter_ns or time.time_ns """
		self.ns: Callable[[], int] = time.perf_counter_ns if perf_counter else time.time_ns
		""" Time function to use """
		self.start_ns: int = 0
		""" Start time in nanoseconds """

	def __enter__(self) -> MeasureTime:
		""" Enter context manager, record start time """
		self.start_ns = self.ns()
		return self

	def __exit__(self, exc_type: type[BaseException]|None, exc_val: BaseException|None, exc_tb: Any|None) -> None:
		""" Exit context manager, calculate duration and print """
		# Measure the execution time (nanoseconds and seconds)
		total_ns: int = self.ns() - self.start_ns
		total_ms: float = total_ns / 1_000_000
		total_s: float = total_ns / 1_000_000_000

		# Print the execution time (nanoseconds if less than 0.1s, seconds otherwise)
		if total_ms < 100:
			self.print_func(f"{self.message}: {total_ms:.3f}ms ({total_ns}ns)")
		elif total_s < 60:
			self.print_func(f"{self.message}: {(total_s):.5f}s")
		else:
			minutes: int = int(total_s) // 60
			seconds: int = int(total_s) % 60
			if minutes < 60:
				self.print_func(f"{self.message}: {minutes}m {seconds}s")
			else:
				hours: int = minutes // 60
				minutes: int = minutes % 60
				if hours < 24:
					self.print_func(f"{self.message}: {hours}h {minutes}m {seconds}s")
				else:
					days: int = hours // 24
					hours: int = hours % 24
					self.print_func(f"{self.message}: {days}d {hours}h {minutes}m {seconds}s")

	async def __aenter__(self) -> MeasureTime:
		""" Enter async context manager, record start time """
		return self.__enter__()

	async def __aexit__(self, exc_type: type[BaseException]|None, exc_val: BaseException|None, exc_tb: Any|None) -> None:
		""" Exit async context manager, calculate duration and print """
		self.__exit__(exc_type, exc_val, exc_tb)

# Context manager to temporarily silence output
class Muffle(AbstractBothContextManager["Muffle"]):
	""" Context manager that temporarily silences output.
	(No thread-safety guaranteed)

	Alternative to stouputils.decorators.silent()

	Examples:
		>>> with Muffle():
		...     print("This will not be printed")
	"""
	def __init__(self, mute_stderr: bool = False) -> None:
		self.mute_stderr: bool = mute_stderr
		""" Attribute remembering if stderr should be muted """
		self.original_stdout: IO[Any]
		""" Attribute remembering original stdout """
		self.original_stderr: IO[Any]
		""" Attribute remembering original stderr """

	def __enter__(self) -> Muffle:
		""" Enter context manager which redirects stdout and stderr to devnull """
		# Redirect stdout to devnull
		self.original_stdout = sys.stdout
		sys.stdout = open(os.devnull, "w", encoding="utf-8")

		# Redirect stderr to devnull if needed
		if self.mute_stderr:
			self.original_stderr = sys.stderr
			sys.stderr = open(os.devnull, "w", encoding="utf-8")

		# Return self
		return self

	def __exit__(self, exc_type: type[BaseException]|None, exc_val: BaseException|None, exc_tb: Any|None) -> None:
		""" Exit context manager which restores original stdout and stderr """
		# Restore original stdout
		sys.stdout.close()
		sys.stdout = self.original_stdout

		# Restore original stderr if needed
		if self.mute_stderr:
			sys.stderr.close()
			sys.stderr = self.original_stderr

	async def __aenter__(self) -> Muffle:
		""" Enter async context manager which redirects stdout and stderr to devnull """
		return self.__enter__()

	async def __aexit__(self, exc_type: type[BaseException]|None, exc_val: BaseException|None, exc_tb: Any|None) -> None:
		""" Exit async context manager which restores original stdout and stderr """
		self.__exit__(exc_type, exc_val, exc_tb)

# Context manager that does nothing
class DoNothing(AbstractBothContextManager["DoNothing"]):
	""" Context manager that does nothing.

	This is a no-op context manager that can be used as a placeholder
	or for conditional context management.

	Different from contextlib.nullcontext because it handles args and kwargs,
	along with **async** context management.

	Examples:
		>>> with DoNothing():
		...     print("This will be printed normally")
		This will be printed normally

		>>> # Conditional context management
		>>> some_condition = True
		>>> ctx = DoNothing() if some_condition else Muffle()
		>>> with ctx:
		...     print("May or may not be printed depending on condition")
		May or may not be printed depending on condition
	"""
	def __init__(self, *args: Any, **kwargs: Any) -> None:
		""" No initialization needed, this is a no-op context manager """
		pass

	def __enter__(self) -> DoNothing:
		""" Enter context manager (does nothing) """
		return self

	def __exit__(self, exc_type: Any, exc_val: Any, exc_tb: Any) -> None:
		""" Exit context manager (does nothing) """
		pass

	async def __aenter__(self) -> DoNothing:
		""" Enter async context manager (does nothing) """
		return self

	async def __aexit__(self, *excinfo: Any) -> None:
		""" Exit async context manager (does nothing) """
		pass
NullContextManager = DoNothing
""" Alias for DoNothing context manager """

# Context manager to temporarily set multiprocessing start method
class SetMPStartMethod(AbstractBothContextManager["SetMPStartMethod"]):
	""" Context manager to temporarily set multiprocessing start method.

	This context manager allows you to temporarily change the multiprocessing start method
	and automatically restores the original method when exiting the context.

	Args:
		start_method (str): The start method to use: "spawn", "fork", or "forkserver"

	Examples:
		.. code-block:: python

			> import multiprocessing as mp
			> import stouputils as stp
			> # Temporarily use spawn method
			> with stp.SetMPStartMethod("spawn"):
			> ...     # Your multiprocessing code here
			> ...     pass

			> # Original method is automatically restored
	"""
	def __init__(self, start_method: str | None) -> None:
		self.start_method: str | None = start_method
		""" The start method to use """
		self.old_method: str | None = None
		""" The original start method to restore """

	def __enter__(self) -> SetMPStartMethod:
		""" Enter context manager which sets the start method """
		if self.start_method is None:
			return self
		import multiprocessing as mp

		self.old_method = mp.get_start_method(allow_none=True)
		if self.old_method != self.start_method:
			mp.set_start_method(self.start_method, force=True)
		return self

	def __exit__(self, exc_type: Any, exc_val: Any, exc_tb: Any) -> None:
		""" Exit context manager which restores the original start method """
		if self.start_method is None:
			return
		import multiprocessing as mp

		if self.old_method != self.start_method:
			mp.set_start_method(self.old_method, force=True)

	async def __aenter__(self) -> SetMPStartMethod:
		""" Enter async context manager which sets the start method """
		return self.__enter__()

	async def __aexit__(self, exc_type: Any, exc_val: Any, exc_tb: Any) -> None:
		""" Exit async context manager which restores the original start method """
		self.__exit__(exc_type, exc_val, exc_tb)




"""
This module contains the GroupingStrategy class, which provides a strategy for grouping images when loading a dataset.

There are 3 strategies, NONE, SIMPLE and CONCATENATE.
Refer to the docstrings of the GroupingStrategy class for more information.
"""
# pyright: reportUnknownMemberType=false

# Imports
from __future__ import annotations

import os
from enum import Enum
from typing import Any

import numpy as np
from numpy.typing import NDArray

from ...decorators import handle_error
from ...parallel import multiprocessing
from ...print import warning
from ...io import clean_path
from ..config.get import DataScienceConfig
from .image_loader import load_images_from_directory
from .xy_tuple import XyTuple


# Grouping strategy class for the dataset
class GroupingStrategy(Enum):
	""" Grouping strategy for the dataset """

	NONE = 0
	""" Default behavior: A subfolder "subject1" is a group of images, all images are grouped together (list of features)
	and the label is the class of the folder above (class1)

	Example file tree:

	- dataset/class1/subject1/image1.png
	- dataset/class1/subject1/image2.png
	- dataset/class1/subject1/image3.png

	Example data (if binary classification):

	- features = [features_image1, features_image2, features_image3] where
		features_image1, features_image2, features_image3 are NDArray[Any] of shape `(224, 224, 3)`
	- labels = [1.0, 0.0]

	If subjects do not have the same number of images,
	the missing images are padded with zeros so every features have the same shape.

	This strategy preserves the relationship between images of the same subject when splitting the dataset,
	ensuring that all images from the same subject stay together in either train or test sets.
	"""

	CONCATENATE = 1
	""" A subfolder "subject1" is a group of images, all images are concatenated into a single feature (NDArray[Any])
	and the label is the class of the folder above (class1)

	Example file tree:

	- dataset/class1/subject1/image1.png
	- dataset/class1/subject1/image2.png
	- dataset/class1/subject1/image3.png

	Example data (if binary classification):

	- features will have a shape of `(224, 224, 3*num_images)` (if RGB images).
		Notice that the concatenation is done along the last axis.
	- labels = [1.0, 0.0]

	If subjects do not have the same number of images,
	the missing images are padded with zeros so every features have the same shape.
	"""

	@staticmethod
	def _load_folder(
		folder_path: str,
		class_idx: int,
		num_classes: int,
		kwargs: dict[str, Any]
	) -> tuple[list[NDArray[Any]], NDArray[Any], tuple[str, ...]]:
		""" Load images from a single folder.

		Args:
			folder_path        (str):               Path to the folder
			class_idx          (int):               Index of the class
			num_classes        (int):               Total number of classes
			kwargs             (dict[str, Any]):    Additional arguments for image_dataset_from_directory
		Returns:
			list[tuple[NDArray[Any], NDArray[Any], str]]: List of tuples containing (images, one-hot label, filepaths)

		Examples:
			.. code-block:: python

				> data = GroupingStrategy._load_folder(
					folder_path="data/pizza/pizza1",
					class_idx=0,
					num_classes=2,
					kwargs={"color_mode": "grayscale"}
				)
				> features, label, filepaths = zip(*data, strict=True)
		"""
		# Load images from the folder
		images_and_paths: list[tuple[NDArray[Any], str]] = load_images_from_directory(folder_path, **kwargs)
		images, paths = zip(*images_and_paths, strict=True) if images_and_paths else ([], [])
		images: list[NDArray[Any]]
		paths: list[str]

		# Create a one-hot encoded label vector
		label: NDArray[Any] = np.zeros(num_classes)
		label[class_idx] = 1.0

		return list(images), label, tuple(paths)

	@staticmethod
	@handle_error(error_log=DataScienceConfig.ERROR_LOG)
	def image_dataset_from_directory(
		grouping_strategy: GroupingStrategy,
		path: str,
		seed: int,
		**kwargs: Any
	) -> tuple[XyTuple, tuple[str, ...], GroupingStrategy]:
		""" Load images from a directory while keeping groups of images together.

		Args:
			grouping_strategy  (GroupingStrategy):  Grouping strategy to use
			path               (str):               Path to the dataset directory
			seed               (int):               Random seed for shuffling
			**kwargs           (Any):               Additional arguments passed to image_dataset_from_directory

		Returns:
			XyTuple:          XyTuple with the data
			tuple[str, ...]:  List of class labels (strings)
			GroupingStrategy: Grouping strategy used (because it can be updated)

		Examples:
			.. code-block:: python

				> data = GroupingStrategy.image_dataset_from_directory(
					grouping_strategy=GroupingStrategy.NONE,
					path="data/pizza",
					seed=42,
					color_mode="grayscale"
				)
				> all_data: XyTuple = data[0]
				> all_labels: tuple[str, ...] = data[1]
		"""
		# Get all subdirectories (classes)
		path = clean_path(path)
		class_dirs: tuple[str, ...] = tuple(d for d in os.listdir(path) if os.path.isdir(f"{path}/{d}"))

		# Check if there are subfolders in each class
		any_subfolders: bool = any(
			os.path.isdir(f"{path}/{class_dir}/{sub_dir}")
			for class_dir in class_dirs for sub_dir in os.listdir(f"{path}/{class_dir}")
		)

		# Verify if wrong grouping strategy, then adjust it
		if grouping_strategy != GroupingStrategy.NONE and not any_subfolders:
			warning(
				f"Strategy is {grouping_strategy.name} but there are no subfolders in each class, adjusting to NONE "
				"as there is no way to group the images together, that just doesn't make sense"
			)
			grouping_strategy = GroupingStrategy.NONE

		# Prepare multithreading arguments
		queue: list[tuple[str, int, int, dict[str, Any]]] = []
		for class_idx, class_dir in enumerate(class_dirs):
			class_path: str = f"{path}/{class_dir}"

			# Get subfolders (class1/subject1/image1.png) to the queue
			sub_folders: list[str] = [d for d in os.listdir(class_path) if os.path.isdir(f"{class_path}/{d}")]
			for sub_folder in sub_folders:
				folder_path: str = f"{class_path}/{sub_folder}"
				queue.append((folder_path, class_idx, len(class_dirs), kwargs))

			# Get files in the class folder
			files: list[str] = [f for f in os.listdir(class_path) if os.path.isfile(f"{class_path}/{f}")]
			for file in files:
				queue.append((f"{class_path}/{file}", class_idx, len(class_dirs), kwargs))

		# Process folders in parallel
		splitted: list[str] = path.split('/')
		description: str = f".../{splitted[-1]}" if len(splitted) > 2 else path
		extracted_folders: list[tuple[list[NDArray[Any]], NDArray[Any], tuple[str, ...]]] = multiprocessing(
			GroupingStrategy._load_folder,
			queue,
			use_starmap=True,
			desc=f"Loading dataset '{description}'"
		)

		# Extract results properly
		all_X: list[list[NDArray[Any]]] = []
		all_y: list[NDArray[Any]] = []
		all_filenames: list[tuple[str, ...]] = []

		# For each folder extracted (each subject maybe)
		for images, label, filepaths in extracted_folders:
			if not images:
				continue	# Skip if no images are found

			to_append_X: list[NDArray[Any]] = []
			to_append_filepaths: list[str] = []

			# For each image of the subject,
			for image, filepath in zip(images, filepaths, strict=True):

				# Add the data
				to_append_X.append(image)
				to_append_filepaths.append(filepath)

			# Append the subject if there are images
			if to_append_X:

				# If concatenate strategy, combine images along the channel axis
				if grouping_strategy == GroupingStrategy.CONCATENATE:
					# Step 1: Make an array of shape (num_images, height, width, channels)
					images_array = np.array(to_append_X)

					# Step 2: Transpose to move channels next to num_images
					# From (num_images, height, width, channels) to (height, width, num_images, channels)
					images_array = np.transpose(images_array, (1, 2, 0, 3))

					# Step 3: Reshape to combine num_images and channels dimensions
					# From (height, width, num_images, channels) to (height, width, num_images * channels)
					images_array = images_array.reshape(images_array.shape[0], images_array.shape[1], -1)

					# Step 4: Add single concatenated feature array
					all_X.append([images_array])

				# Else, just add the images
				else:
					all_X.append(to_append_X)

				all_y.append(label)
				all_filenames.append(tuple(to_append_filepaths))

		# Fix different sizes of images
		if grouping_strategy == GroupingStrategy.CONCATENATE:
			all_X = GroupingStrategy.fix_different_sizes(all_X, grouping_strategy)

		# Shuffle the data
		combined = list(zip(all_X, all_y, all_filenames, strict=True))
		np.random.seed(seed)
		np.random.shuffle(combined) # pyright: ignore [reportArgumentType]
		all_X, all_y, all_filenames = zip(*combined, strict=True)

		# Create a XyTuple and return it
		return XyTuple(all_X, all_y, tuple(all_filenames)), class_dirs, grouping_strategy


	@staticmethod
	def fix_different_sizes(data: list[list[NDArray[Any]]], grouping_strategy: GroupingStrategy) -> list[list[NDArray[Any]]]:
		""" Fix different sizes of images in a list of lists of numpy arrays.

		Simple strategy will add empty images to shape[0]
		Concatenate strategy will add empty channels to shape[-1]

		Args:
			data                (list[list[NDArray[Any]]]):  List of lists of numpy arrays
			grouping_strategy   (GroupingStrategy):        Grouping strategy used

		Returns:
			list[list[NDArray[Any]]]:  List of lists of numpy arrays with consistent shapes

		Examples:
			>>> # Concatenate grouping strategy
			>>> data = [[np.zeros((7, 224, 224, 3))], [np.zeros((1, 224, 224, 1))]]
			>>> data = GroupingStrategy.fix_different_sizes(data, GroupingStrategy.CONCATENATE)
			>>> data[0][0].shape
			(7, 224, 224, 3)
			>>> data[1][0].shape
			(1, 224, 224, 3)
			>>> data[1][0].shape[0] == data[0][0].shape[0]
			False
			>>> data[1][0].shape[-1] == data[0][0].shape[-1]
			True
		"""
		# Add empty channels to images that have less channels than others
		if grouping_strategy == GroupingStrategy.CONCATENATE:
			# Find the maximum number of channels across all images in all groups
			max_num_channels: int = max(x.shape[-1] for group in data for x in group)

			for i, group in enumerate(data):
				for j, image in enumerate(group):
					if image.shape[-1] < max_num_channels:
						# Calculate how many times to repeat the channels
						repeat_count: int = int(np.ceil(max_num_channels / image.shape[-1]))

						# Repeat the channels and then slice to get exactly the right number
						repeated_channels = np.repeat(image, repeat_count, axis=-1)
						data[i][j] = repeated_channels[..., :max_num_channels]

		# Return the fixed list of lists of numpy arrays
		return data





# Imports
from .auto_contrast import auto_contrast_image
from .axis_flip import flip_image
from .bias_field_correction import bias_field_correction_image
from .binary_threshold import binary_threshold_image
from .blur import blur_image
from .brightness import brightness_image
from .canny import canny_image
from .clahe import clahe_image
from .contrast import contrast_image
from .curvature_flow_filter import curvature_flow_filter_image
from .denoise import (
    adaptive_denoise_image,
    bilateral_denoise_image,
    nlm_denoise_image,
    tv_denoise_image,
    wavelet_denoise_image,
)
from .invert import invert_image
from .laplacian import laplacian_image
from .median_blur import median_blur_image
from .noise import noise_image
from .normalize import normalize_image
from .random_erase import random_erase_image
from .resize import resize_image
from .rotation import rotate_image
from .salt_pepper import salt_pepper_image
from .sharpening import sharpen_image
from .shearing import shear_image
from .threshold import threshold_image
from .translation import translate_image
from .zoom import zoom_image

__all__ = [
    "adaptive_denoise_image",
    "auto_contrast_image",
    "bias_field_correction_image",
    "bilateral_denoise_image",
    "binary_threshold_image",
    "blur_image",
    "brightness_image",
    "canny_image",
    "clahe_image",
    "contrast_image",
    "curvature_flow_filter_image",
    "flip_image",
    "invert_image",
    "laplacian_image",
    "median_blur_image",
    "nlm_denoise_image",
    "noise_image",
    "normalize_image",
    "random_erase_image",
    "resize_image",
    "rotate_image",
    "salt_pepper_image",
    "sharpen_image",
    "shear_image",
    "threshold_image",
    "translate_image",
    "tv_denoise_image",
    "wavelet_denoise_image",
    "zoom_image",
]





# pyright: reportMissingTypeStubs=false
# pyright: reportUnknownMemberType=false

# Imports
from typing import Any

from keras.callbacks import ModelCheckpoint


class ModelCheckpointV2(ModelCheckpoint):
	""" Model checkpoint callback but only starts after a given number of epochs.

	Args:
		epochs_before_start (int): Number of epochs before starting the checkpointing
	"""

	def __init__(self, epochs_before_start: int = 3, *args: Any, **kwargs: Any) -> None:
		super().__init__(*args, **kwargs)
		self.epochs_before_start = epochs_before_start
		self.current_epoch = 0

	def on_batch_end(self, batch: int, logs: dict[str, Any] | None = None) -> None:
		if self.current_epoch >= self.epochs_before_start:
			super().on_batch_end(batch, logs)

	def on_epoch_end(self, epoch: int, logs: dict[str, Any] | None = None) -> None:
		self.current_epoch = epoch
		if epoch >= self.epochs_before_start:
			super().on_epoch_end(epoch, logs)




WAIFU2X_NCNN_VULKAN_RELEASES: dict[str, str]
FFMPEG_RELEASES: dict[str, str]
YOUTUBE_BITRATE_RECOMMENDATIONS: dict[str, dict[str, dict[int, int]]]

class Config:
    """ Configuration class for the upscaler. """
    JPG_QUALITY: int
    VIDEO_FINAL_BITRATE: int
    FFMPEG_EXECUTABLE: str
    FFMPEG_ARGS: tuple[str, ...]
    FFPROBE_EXECUTABLE: str
    FFMPEG_CHECK_HELP_TEXT: str
    UPSCALER_EXECUTABLE: str
    UPSCALER_ARGS: tuple[str, ...]
    UPSCALER_EXECUTABLE_HELP_TEXT: str
    SLIGHTLY_FASTER_MODE: bool
    upscaler_executable_checked: bool
    ffmpeg_executable_checked: bool



"""
This module provides utilities for collection manipulation:

- unique_list: Remove duplicates from a list while preserving order using object id, hash or str
- sort_dict_keys: Sort dictionary keys using a given order list (ascending or descending)
- upsert_in_dataframe: Insert or update a row in a Polars DataFrame based on primary keys
- array_to_disk: Easily handle large numpy arrays on disk using zarr for efficient storage and access.

.. image:: https://raw.githubusercontent.com/Stoupy51/stouputils/refs/heads/main/assets/collections_module.gif
  :alt: stouputils collections examples
"""

# Imports
import atexit
import os
import shutil
import tempfile
from collections.abc import Iterable
from typing import TYPE_CHECKING, Any, Literal, TypeVar

# Lazy imports for typing
if TYPE_CHECKING:
	import numpy as np
	import polars as pl
	import zarr  # pyright: ignore[reportMissingTypeStubs]
	from numpy.typing import NDArray

# Typing
T = TypeVar("T")

# Functions
def unique_list[T](list_to_clean: Iterable[T], method: Literal["id", "hash", "str"] = "str") -> list[T]:
	""" Remove duplicates from the list while keeping the order using ids (default) or hash or str

	Args:
		list_to_clean	(Iterable[T]):					The list to clean
		method			(Literal["id", "hash", "str"]):	The method to use to identify duplicates
	Returns:
		list[T]: The cleaned list

	Examples:
		>>> unique_list([1, 2, 3, 2, 1], method="id")
		[1, 2, 3]

		>>> s1 = {1, 2, 3}
		>>> s2 = {2, 3, 4}
		>>> s3 = {1, 2, 3}
		>>> unique_list([s1, s2, s1, s1, s3, s2, s3], method="id")
		[{1, 2, 3}, {2, 3, 4}, {1, 2, 3}]

		>>> s1 = {1, 2, 3}
		>>> s2 = {2, 3, 4}
		>>> s3 = {1, 2, 3}
		>>> unique_list([s1, s2, s1, s1, s3, s2, s3], method="str")
		[{1, 2, 3}, {2, 3, 4}]
	"""
	# Initialize the seen ids set and the result list
	seen: set[int | str] = set()
	result: list[T] = []

	# Iterate over each item in the list
	for item in list_to_clean:
		if method == "id":
			item_identifier = id(item)
		elif method == "hash":
			item_identifier = hash(item)
		elif method == "str":
			item_identifier = str(item)
		else:
			raise ValueError(f"Invalid method: {method}")

		# If the item id is not in the seen ids set, add it to the seen ids set and append the item to the result list
		if item_identifier not in seen:
			seen.add(item_identifier)
			result.append(item)

	# Return the cleaned list
	return result

def sort_dict_keys[T](dictionary: dict[T, Any], order: list[T], reverse: bool = False) -> dict[T, Any]:
	""" Sort dictionary keys using a given order list (reverse optional)

	Args:
		dictionary	(dict[T, Any]):	The dictionary to sort
		order		(list[T]):		The order list
		reverse		(bool):			Whether to sort in reverse order (given to sorted function which behaves differently than order.reverse())
	Returns:
		dict[T, Any]: The sorted dictionary

	Examples:
		>>> sort_dict_keys({'b': 2, 'a': 1, 'c': 3}, order=["a", "b", "c"])
		{'a': 1, 'b': 2, 'c': 3}

		>>> sort_dict_keys({'b': 2, 'a': 1, 'c': 3}, order=["a", "b", "c"], reverse=True)
		{'c': 3, 'b': 2, 'a': 1}

		>>> sort_dict_keys({'b': 2, 'a': 1, 'c': 3, 'd': 4}, order=["c", "b"])
		{'c': 3, 'b': 2, 'a': 1, 'd': 4}
	"""
	return dict(sorted(dictionary.items(), key=lambda x: order.index(x[0]) if x[0] in order else len(order), reverse=reverse))

def upsert_in_dataframe(
	df: "pl.DataFrame",
	new_entry: dict[str, Any],
	primary_keys: dict[str, Any] | None = None
) -> "pl.DataFrame":
	""" Insert or update a row in the Polars DataFrame based on primary keys.

	Args:
		df				(pl.DataFrame):		The Polars DataFrame to update.
		new_entry		(dict[str, Any]):	The new entry to insert or update.
		primary_keys	(dict[str, Any]):	The primary keys to identify the row (default: empty).
	Returns:
		pl.DataFrame: The updated Polars DataFrame.
	"""
	# Imports
	import polars as pl

	# Create new DataFrame if file doesn't exist or is invalid
	if df.is_empty():
		return pl.DataFrame([new_entry])

	# If no primary keys provided, return DataFrame with new entry appended
	if not primary_keys:
		new_row_df = pl.DataFrame([new_entry])
		return pl.concat([df, new_row_df], how="diagonal_relaxed")

	# Build mask based on primary keys
	mask: pl.Expr = pl.lit(True)
	for key, value in primary_keys.items():
		if key in df.columns:
			mask = mask & (df[key] == value)
		else:
			# Primary key column doesn't exist, so no match possible
			mask = pl.lit(False)
			break

	# Insert or update row based on primary keys
	if df.select(mask).to_series().any():
		# Update existing row
		for key, value in new_entry.items():
			if key in df.columns:
				df = df.with_columns(pl.when(mask).then(pl.lit(value)).otherwise(pl.col(key)).alias(key))
			else:
				# Add new column if it doesn't exist
				df = df.with_columns(pl.when(mask).then(pl.lit(value)).otherwise(None).alias(key))
		return df
	else:
		# Insert new row
		new_row_df = pl.DataFrame([new_entry])
		return pl.concat([df, new_row_df], how="diagonal_relaxed")

def array_to_disk(
	data: "NDArray[Any] | zarr.Array",
	delete_input: bool = True,
	more_data: "NDArray[Any] | zarr.Array | None" = None
) -> tuple["zarr.Array", str, int]:
	""" Easily handle large numpy arrays on disk using zarr for efficient storage and access.

	Zarr provides a simpler and more efficient alternative to np.memmap with better compression
	and chunking capabilities.

	Args:
		data			(NDArray | zarr.Array):	The data to save/load as a zarr array
		delete_input	(bool):	Whether to delete the input data after creating the zarr array
		more_data		(NDArray | zarr.Array | None): Additional data to append to the zarr array
	Returns:
		tuple[zarr.Array, str, int]: The zarr array, the directory path, and the total size in bytes

	Examples:
		>>> import numpy as np
		>>> data = np.random.rand(1000, 1000)
		>>> zarr_array = array_to_disk(data)[0]
		>>> zarr_array.shape
		(1000, 1000)

		>>> more_data = np.random.rand(500, 1000)
		>>> longer_array, dir_path, total_size = array_to_disk(zarr_array, more_data=more_data)
	"""
	def dir_size(directory: str) -> int:
		return sum(
			os.path.getsize(os.path.join(dirpath, filename))
			for dirpath, _, filenames in os.walk(directory)
			for filename in filenames
		)

	# Imports
	try:
		import zarr  # pyright: ignore[reportMissingTypeStubs]
	except ImportError as e:
		raise ImportError("zarr is required for array_to_disk function. Please install it via 'pip install zarr'.") from e

	# If data is already a zarr.Array and more_data is present, just append and return
	if isinstance(data, zarr.Array) and more_data is not None:
		original_size: int = data.shape[0]
		new_shape: tuple[int, ...] = (original_size + more_data.shape[0], *data.shape[1:])
		data.resize(new_shape)
		data[original_size:] = more_data[:]

		# Delete more_data if specified, calculate size, and return
		if delete_input:
			del more_data
		store_path: str = str(data.store.path if hasattr(data.store, 'path') else data.store) # type: ignore
		return data, store_path, dir_size(store_path)

	# Create a temporary directory to store the zarr array (with compression (auto-chunking for optimal performance))
	temp_dir: str = tempfile.mkdtemp()
	zarr_array: zarr.Array = zarr.open_array(temp_dir, mode="w", shape=data.shape, dtype=data.dtype, chunks=True) # pyright: ignore[reportUnknownMemberType]
	zarr_array[:] = data[:]

	# If additional data is provided, resize and append
	if more_data is not None:
		original_size = data.shape[0]
		new_shape = (original_size + more_data.shape[0], *data.shape[1:])
		zarr_array.resize(new_shape)
		zarr_array[original_size:] = more_data[:]

	# Delete the original data from memory if specified
	if delete_input:
		del data
		if more_data is not None:
			del more_data

	# Register a cleanup function to delete the zarr directory at exit
	atexit.register(lambda: shutil.rmtree(temp_dir, ignore_errors=True))

	# Return all
	return zarr_array, temp_dir, dir_size(temp_dir)

if __name__ == "__main__":

	# Example usage of array_to_disk (now using zarr)
	print("\nZarr Example:")
	data = np.random.rand(1000, 1000)
	zarr_array, dir_path, total_size = array_to_disk(data, delete_input=True)
	print(f"Zarr array shape: {zarr_array.shape}, directory: {dir_path}, size: {total_size:,} bytes")
	print(f"Compression ratio: {(data.nbytes / total_size):.2f}x")

	# Make it longer (1000x1000 -> 1500x1000)
	data2 = np.random.rand(500, 1000)
	longer_array, dir_path, total_size = array_to_disk(zarr_array, more_data=data2)
	print(f"\nLonger zarr array shape: {longer_array.shape}, directory: {dir_path}, size: {total_size:,} bytes")
	print(f"Compression ratio: {(1500 * 1000 * 8 / total_size):.2f}x")




import numpy as np
from .io import super_open as super_open
from .print import debug as debug, info as info
from PIL import Image
from collections.abc import Callable
from numpy.typing import NDArray
from typing import Any, TypeVar

PIL_Image_or_NDArray = TypeVar('PIL_Image_or_NDArray', bound='Image.Image | NDArray[np.number]')

def image_resize[PIL_Image_or_NDArray](image: PIL_Image_or_NDArray, max_result_size: int, resampling: Image.Resampling | None = None, min_or_max: Callable[[int, int], int] = ..., return_type: type[PIL_Image_or_NDArray] | str = 'same', keep_aspect_ratio: bool = True) -> Any:
    ''' Resize an image while preserving its aspect ratio by default.
\tScales the image so that its largest dimension equals max_result_size.

\tArgs:
\t\timage             (Image.Image | np.ndarray): The image to resize.
\t\tmax_result_size   (int):                      Maximum size for the largest dimension.
\t\tresampling        (Image.Resampling | None):  PIL resampling filter to use (default: Image.Resampling.LANCZOS).
\t\tmin_or_max        (Callable):                 Function to use to get the minimum or maximum of the two ratios.
\t\treturn_type       (type | str):               Type of the return value (Image.Image, np.ndarray, or "same" to match input type).
\t\tkeep_aspect_ratio (bool):                     Whether to keep the aspect ratio.
\tReturns:
\t\tImage.Image | NDArray[np.number]: The resized image with preserved aspect ratio.
\tExamples:
\t\t>>> # Test with (height x width x channels) numpy array
\t\t>>> import numpy as np
\t\t>>> array = np.random.randint(0, 255, (100, 50, 3), dtype=np.uint8)
\t\t>>> image_resize(array, 100).shape
\t\t(100, 50, 3)
\t\t>>> image_resize(array, 100, min_or_max=max).shape
\t\t(100, 50, 3)
\t\t>>> image_resize(array, 100, min_or_max=min).shape
\t\t(200, 100, 3)

\t\t>>> # Test with PIL Image
\t\t>>> from PIL import Image
\t\t>>> pil_image: Image.Image = Image.new(\'RGB\', (200, 100))
\t\t>>> image_resize(pil_image, 50).size
\t\t(50, 25)
\t\t>>> # Test with different return types
\t\t>>> resized_array = image_resize(array, 50, return_type=np.ndarray)
\t\t>>> isinstance(resized_array, np.ndarray)
\t\tTrue
\t\t>>> resized_array.shape
\t\t(50, 25, 3)
\t\t>>> # Test with different resampling methods
\t\t>>> image_resize(pil_image, 50, resampling=Image.Resampling.NEAREST).size
\t\t(50, 25)
\t'''
def auto_crop[PIL_Image_or_NDArray](image: PIL_Image_or_NDArray, mask: NDArray[np.bool_] | None = None, threshold: int | float | Callable[[NDArray[np.number]], int | float] | None = None, return_type: type[PIL_Image_or_NDArray] | str = 'same', contiguous: bool = True) -> Any:
    ''' Automatically crop an image to remove zero or uniform regions.

\tThis function crops the image to keep only the region where pixels are non-zero
\t(or above a threshold). It can work with a mask or directly analyze the image.

\tArgs:
\t\timage       (Image.Image | NDArray):\t  The image to crop.
\t\tmask        (NDArray[bool] | None):       Optional binary mask indicating regions to keep.
\t\tthreshold   (int | float | Callable):     Threshold value or function (default: np.min).
\t\treturn_type (type | str):                 Type of the return value (Image.Image, NDArray[np.number], or "same" to match input type).
\t\tcontiguous  (bool):                       If True (default), crop to bounding box. If False, remove entire rows/columns with no content.
\tReturns:
\t\tImage.Image | NDArray[np.number]: The cropped image.

\tExamples:
\t\t>>> # Test with numpy array with zeros on edges
\t\t>>> import numpy as np
\t\t>>> array = np.zeros((100, 100, 3), dtype=np.uint8)
\t\t>>> array[20:80, 30:70] = 255  # White rectangle in center
\t\t>>> cropped = auto_crop(array, return_type=np.ndarray)
\t\t>>> cropped.shape
\t\t(60, 40, 3)

\t\t>>> # Test with custom mask
\t\t>>> mask = np.zeros((100, 100), dtype=bool)
\t\t>>> mask[10:90, 10:90] = True
\t\t>>> cropped_with_mask = auto_crop(array, mask=mask, return_type=np.ndarray)
\t\t>>> cropped_with_mask.shape
\t\t(80, 80, 3)

\t\t>>> # Test with PIL Image
\t\t>>> from PIL import Image
\t\t>>> pil_image = Image.new(\'RGB\', (100, 100), (0, 0, 0))
\t\t>>> from PIL import ImageDraw
\t\t>>> draw = ImageDraw.Draw(pil_image)
\t\t>>> draw.rectangle([25, 25, 75, 75], fill=(255, 255, 255))
\t\t>>> cropped_pil = auto_crop(pil_image)
\t\t>>> cropped_pil.size
\t\t(51, 51)

\t\t>>> # Test with threshold
\t\t>>> array_gray = np.ones((100, 100), dtype=np.uint8) * 10
\t\t>>> array_gray[20:80, 30:70] = 255
\t\t>>> cropped_threshold = auto_crop(array_gray, threshold=50, return_type=np.ndarray)
\t\t>>> cropped_threshold.shape
\t\t(60, 40)

\t\t>>> # Test with callable threshold (using lambda to avoid min value)
\t\t>>> array_gray2 = np.ones((100, 100), dtype=np.uint8) * 10
\t\t>>> array_gray2[20:80, 30:70] = 255
\t\t>>> cropped_max = auto_crop(array_gray2, threshold=lambda x: 50, return_type=np.ndarray)
\t\t>>> cropped_max.shape
\t\t(60, 40)

\t>>> # Test with non-contiguous crop
\t>>> array_sparse = np.zeros((100, 100, 3), dtype=np.uint8)
\t>>> array_sparse[10, 10] = 255
\t>>> array_sparse[50, 50] = 255
\t>>> array_sparse[90, 90] = 255
\t>>> cropped_contiguous = auto_crop(array_sparse, contiguous=True, return_type=np.ndarray)
\t>>> cropped_contiguous.shape  # Bounding box from (10,10) to (90,90)
\t(81, 81, 3)
\t>>> cropped_non_contiguous = auto_crop(array_sparse, contiguous=False, return_type=np.ndarray)
\t>>> cropped_non_contiguous.shape  # Only rows/cols 10, 50, 90
\t(3, 3, 3)

\t>>> # Test with 3D crop on depth dimension
\t>>> array_3d = np.zeros((50, 50, 10), dtype=np.uint8)
\t>>> array_3d[10:40, 10:40, 2:8] = 255  # Content only in depth slices 2-7
\t>>> cropped_3d = auto_crop(array_3d, contiguous=True, return_type=np.ndarray)
\t>>> cropped_3d.shape  # Should crop all 3 dimensions
\t(30, 30, 6)
\t'''
def numpy_to_gif(path: str, array: NDArray[np.integer | np.floating | np.bool_], duration: int = 100, loop: int = 0, mkdir: bool = True, **kwargs: Any) -> None:
    ''' Generate a \'.gif\' file from a numpy array for 3D/4D visualization.

\tArgs:
\t\tpath     (str):     Path to the output .gif file.
\t\tarray    (NDArray): Numpy array to be dumped (must be 3D or 4D).
\t\t\t3D: (depth, height, width) - e.g. (64, 1024, 1024)
\t\t\t4D: (depth, height, width, channels) - e.g. (50, 64, 1024, 3)
\t\tduration (int):     Duration between frames in milliseconds.
\t\tloop     (int):     Number of loops (0 = infinite).
\t\tmkdir    (bool):    Create the directory if it does not exist.
\t\t**kwargs (Any):     Additional keyword arguments for PIL.Image.save().

\tExamples:

\t\t.. code-block:: python

\t\t\t> # 3D array example
\t\t\t> array = np.random.randint(0, 256, (10, 100, 100), dtype=np.uint8)
\t\t\t> numpy_to_gif("output_10_frames_100x100.gif", array, duration=200, loop=0)

\t\t\t> # 4D array example (batch of 3D images)
\t\t\t> array_4d = np.random.randint(0, 256, (5, 10, 100, 3), dtype=np.uint8)
\t\t\t> numpy_to_gif("output_50_frames_100x100.gif", array_4d, duration=200)

\t\t\t> total_duration = 1000  # 1 second
\t\t\t> numpy_to_gif("output_1s.gif", array, duration=total_duration // len(array))
\t'''
def numpy_to_obj(path: str, array: NDArray[np.integer | np.floating | np.bool_], threshold: float = 0.5, step_size: int = 1, pad_array: bool = True, verbose: int = 0) -> None:
    ''' Generate a \'.obj\' file from a numpy array for 3D visualization using marching cubes.

\tArgs:
\t\tpath      (str):     Path to the output .obj file.
\t\tarray     (NDArray): Numpy array to be dumped (must be 3D).
\t\tthreshold (float):   Threshold level for marching cubes (0.5 for binary data).
\t\tstep_size (int):     Step size for marching cubes (higher = simpler mesh, faster generation).
\t\tpad_array (bool):    If True, pad array with zeros to ensure closed volumes for border cells.
\t\tverbose   (int):     Verbosity level (0 = no output, 1 = some output, 2 = full output).

\tExamples:

\t\t.. code-block:: python

\t\t\t> array = np.random.rand(64, 64, 64) > 0.5  # Binary volume
\t\t\t> numpy_to_obj("output_mesh.obj", array, threshold=0.5, step_size=2, pad_array=True, verbose=1)

\t\t\t> array = my_3d_data  # Some 3D numpy array (e.g. human lung scan)
\t\t\t> numpy_to_obj("output_mesh.obj", array, threshold=0.3)
\t'''




# Imports
import argparse
import os
import sys

from ...decorators import handle_error, measure_time
from ...print import info
from ...parallel import multithreading
from ..config.get import DataScienceConfig
from ..dataset import LOWER_GS
from ..models.all import ALL_MODELS, CLASS_MAP

# Constants
MODEL_HELP: str = "Model to use"
KFOLD_HELP: str = "Number of folds for k-fold cross validation (0 = no k-fold)"
TRANSFER_LEARNING_HELP: str = "Transfer learning source (imagenet, None, \"data/dataset_name\")"
GROUPING_HELP: str = "Grouping strategy for the dataset"
GROUPING_CHOICES: tuple[str, ...] = (*LOWER_GS, "all")
GRID_SEARCH_HELP: str = "If grid search should be performed on hyperparameters"
MAX_WORKERS_HELP: str = "Maximum number of threads for processing"
VERBOSE_HELP: str = "Verbosity level"
PARSER_DESCRIPTION: str = "Command-line interface for exhaustive process."


# Main function
@measure_time(printer=info, message="Total execution time of the script")
@handle_error(exceptions=(KeyboardInterrupt, Exception), error_log=DataScienceConfig.ERROR_LOG)
def exhaustive_process(
	datasets_to_process: list[tuple[str, str]],
	main_script_path: str,

	default_kfold: int = 0,
	default_transfer_learning: str = "imagenet",
	default_grouping: str = "none",
	default_max_workers: int = 1,
	default_verbose: int = 100,
) -> None:
	""" Process all datasets through preprocessing, augmentation, and training.

	This script will:
	1. Verify if the datasets exist
	2. Prepare commands for training models on each dataset
	3. Execute the commands with the specified parameters
	4. Support multiple grouping strategies and model architectures
	5. Allow for k-fold cross-validation and grid search optimization

	Args:
		datasets_to_process (list[tuple[str, str]]): List of dataset paths to process.
			Each tuple contains (dataset_path, based_of_path), e.g. [("aug_preprocessed_path", "preprocessed_path")].
		main_script_path (str): Path to the main script, e.g. "src/main.py"
		default_model (str): Default model architecture to use for training.
		default_kfold (int): Default number of folds for k-fold cross validation.
		default_transfer_learning (str): Default source for transfer learning.
		default_grouping_strategy (str): Default strategy for grouping dataset images.
		default_max_workers (int): Default maximum number of threads for processing.
		default_verbose (int): Default verbosity level for training output.

	Returns:
		None: This function does not return anything.
	"""
	info("Starting the script...")

	# Parse the arguments
	parser = argparse.ArgumentParser(description=PARSER_DESCRIPTION)
	parser.add_argument("--model", type=str, choices=ALL_MODELS, help=MODEL_HELP)
	parser.add_argument("--kfold", type=int, default=default_kfold, help=KFOLD_HELP)
	parser.add_argument("--transfer_learning", type=str, default=default_transfer_learning, help=TRANSFER_LEARNING_HELP)
	parser.add_argument("--grouping_strategy", type=str, default=default_grouping, choices=GROUPING_CHOICES, help=GROUPING_HELP)
	parser.add_argument("--grid_search", action="store_true", help=GRID_SEARCH_HELP)
	parser.add_argument("--max_workers", type=int, default=default_max_workers, help=MAX_WORKERS_HELP)
	parser.add_argument("--verbose", type=int, default=default_verbose, help=VERBOSE_HELP)
	args: argparse.Namespace = parser.parse_args()

	# Extract more arguments
	grouping_strategies: tuple[str, ...] = LOWER_GS if args.grouping_strategy == "all" else (args.grouping_strategy,)

	# Step 1: Verify if the datasets exist
	for dataset_path, based_of in datasets_to_process:
		if not os.path.exists(dataset_path):
			raise FileNotFoundError(f"Dataset not found: '{dataset_path}'")
		if based_of and not os.path.exists(based_of):
			raise FileNotFoundError(f"Based of dataset not found: '{based_of}'")


	# Step 2: Prepare all commands
	commands: list[str] = []
	for dataset_path, based_of in datasets_to_process:
		for grouping_strategy in grouping_strategies:
			info(f"Training on dataset: {dataset_path}")
			based_of_arg: str = f"--based_of {based_of} " if based_of else ""
			grid_search_arg: str = "--grid_search " if args.grid_search else ""

			# Iterate over each model in ROUTINE_MAP
			for model_names in CLASS_MAP.values():

				# Check if the model is in the list of model names
				if args.model in model_names:

					# Get the model name from the list of model names
					# Ex: "good" is in ("densenet121", "densenets", "all", "good"), we take the first one: "densenet121"
					model_name: str = model_names[0]

					# Build base command
					base_cmd: str = (
						f"{sys.executable} {main_script_path} "
						f"--model {model_name} "
						f"--verbose {args.verbose} "
						f"--input {dataset_path} "
						f"--transfer_learning {args.transfer_learning} "
						f"--grouping_strategy {grouping_strategy} "
						f"{based_of_arg}"
						f"{grid_search_arg}"
					)

					# Single run with or without k-fold based on args.kfold
					kfold_arg: str = f"--kfold {args.kfold}" if args.kfold != 0 else ""
					commands.append(f"{base_cmd} {kfold_arg}")

	# Run all commands
	def runner(cmd: str) -> None:
		info(f"Executing command: '{cmd}'")
		sys.stdout.flush()
		sys.stderr.flush()
		os.system(cmd)
	multithreading(
		runner,
		commands,
		desc="Processing all datasets",
		max_workers=args.max_workers,
		delay_first_calls=2.0
	)




from ..io import super_open as super_open
from typing import Any

def read_pyproject(pyproject_path: str) -> dict[str, Any]:
    ''' Read the pyproject.toml file.

\tArgs:
\t\tpyproject_path: Path to the pyproject.toml file.
\tReturns:
\t\tdict[str, Any]: The content of the pyproject.toml file.
\tExample:
\t\t>>> content = read_pyproject("pyproject.toml")
\t\t>>> "." in content["project"]["version"]
\t\tTrue
\t'''
def format_toml_lists(content: str) -> str:
    ''' Format TOML lists with indentation.

\tArgs:
\t\tcontent (str): The content of the pyproject.toml file.
\tReturns:
\t\tstr: The formatted content with properly indented lists.
\tExample:
\t\t>>> toml_content = \'\'\'[project]
\t\t... dependencies = [ "tqdm>=4.0.0", "requests>=2.20.0", "pyyaml>=6.0.0", ]\'\'\'
\t\t>>> format_toml_lists(toml_content).replace("\\t", "    ") == \'\'\'[project]
\t\t... dependencies = [
\t\t...     "tqdm>=4.0.0",
\t\t...     "requests>=2.20.0",
\t\t...     "pyyaml>=6.0.0",
\t\t... ]\'\'\'
\t\tTrue
\t'''
def write_pyproject(path: str, content: dict[str, Any]) -> None:
    """ Write to the pyproject.toml file with properly indented lists.

\tArgs:
\t\tpath: Path to the pyproject.toml file.
\t\tcontent: Content to write to the pyproject.toml file.
\t"""
def increment_version_from_input(version: str) -> str:
    ''' Increment the version.

\tArgs:
\t\tversion: The version to increment. (ex: "0.1.0")
\tReturns:
\t\tstr: The incremented version. (ex: "0.1.1")
\tExample:
\t\t>>> increment_version_from_input("0.1.0")
\t\t\'0.1.1\'
\t\t>>> increment_version_from_input("1.2.9")
\t\t\'1.2.10\'
\t'''
def increment_version_from_pyproject(path: str) -> None:
    """ Increment the version in the pyproject.toml file.

\tArgs:
\t\tpath: Path to the pyproject.toml file.
\t"""
def get_version_from_pyproject(path: str) -> str:
    ''' Get the version from the pyproject.toml file.

\tArgs:
\t\tpath: Path to the pyproject.toml file.
\tReturns:
\t\tstr: The version. (ex: "0.1.0")
\t'''



""" This module contains utilities for generating stub files using stubgen.

- generate_stubs: Generate stub files for a Python package using stubgen
- stubs_full_routine: Generate stub files for a Python package

"""

# Imports
import os
from collections.abc import Callable

from ..decorators import LogLevels, handle_error


def generate_stubs(
	package_name: str,
	extra_args: str = "--include-docstrings --include-private",
) -> int:
	""" Generate stub files for a Python package using stubgen.

	Note: stubgen generates stubs in the 'out' directory by default in the current working directory.

	Args:
		package_name  (str): Name of the package to generate stubs for.
		extra_args    (str): Extra arguments to pass to stubgen. Defaults to "--include-docstrings --include-private".
	Returns:
		int: Return code of the os.system call.
	"""
	try:
		from mypy.stubgen import main as stubgen_main
	except ImportError as e:
		raise ImportError("mypy is required for array_to_disk function. Please install it via 'pip install mypy'.") from e
	try:
		stubgen_main(["-p", package_name, *extra_args.split()])
		return 0
	except Exception:
		return 1

def clean_stubs_directory(output_directory: str, package_name: str) -> None:
	""" Clean the stubs directory by deleting all .pyi files.

	Args:
		output_directory  (str): Directory to clean.
		package_name      (str): Package name subdirectory. Only cleans output_directory/package_name.
	"""
	target_dir: str = os.path.join(output_directory, package_name)
	if os.path.exists(target_dir):
		for root, _, files in os.walk(target_dir):
			for file in files:
				if file.endswith(".pyi"):
					os.remove(os.path.join(root, file))

@handle_error(message="Error while doing the stubs full routine", error_log=LogLevels.ERROR_TRACEBACK)
def stubs_full_routine(
	package_name: str,
	output_directory: str = "typings",
	extra_args: str = "--include-docstrings --include-private",
	clean_before: bool = False,

	generate_stubs_function: Callable[[str, str], int] = generate_stubs,
	clean_stubs_function: Callable[[str, str], None] = clean_stubs_directory,
) -> None:
	""" Generate stub files for a Python package using stubgen.

	Note: stubgen generates stubs in the 'out' directory by default in the current working directory.

	Args:
		package_name              (str):                       Name of the package to generate stubs for.
		output_directory          (str):                       Directory to clean before generating stubs. Defaults to "typings".
			This parameter is used for cleaning the directory before stub generation.
		extra_args                (str):                       Extra arguments to pass to stubgen. Defaults to "--include-docstrings --include-private".
		clean_before              (bool):                      Whether to clean the output directory before generating stubs. Defaults to False.
		generate_stubs_function   (Callable[[str, str], int]): Function to generate stubs.
			Defaults to :func:`generate_stubs`.
		clean_stubs_function      (Callable[[str], None]):     Function to clean the stubs directory.
			Defaults to :func:`clean_stubs_directory`.
	Raises:
		Exception: If stub generation fails.
	"""
	if clean_before:
		clean_stubs_function(output_directory, package_name)
	extra_args += f" -o {output_directory}"

	if generate_stubs_function(package_name, extra_args) != 0:
		raise Exception(f"Error while generating stubs for {package_name}")

